import pathlib
import torch
import pickle
import numpy as np
import gzip
import json
import importlib
import shutil
from typing import Union

PathLike = Union[str, pathlib.Path]


class Obj:
    name = None

    def __init__(self, name=None):
        self.name = name


def is_obj(x):
    return isinstance(x, Obj)


class Flowable(type):
    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)

        # compile objects
        for attr_id, attr in cls.__dict__.items():
            if is_obj(attr):
                assert isinstance(attr_id, str)
                attr.name = attr_id


class Flow(metaclass=Flowable):
    """
    A folder on disk that can contain other folders or objects
    """

    path: pathlib.Path
    default_name = None

    def __init__(self, path=None, folder=None, name=None, reset=False):
        if isinstance(path, str):
            path = pathlib.Path(path)
        if path is None:
            if folder is None:
                # make temporary
                try:
                    import pathlibfs
                except ImportError:
                    raise ImportError("To create a temporary flow, install pathlibfs")
                from uuid import uuid4

                name = str(uuid4())
                folder = pathlibfs.Path("memory://")
            elif name is None:
                if self.default_name is None:
                    raise ValueError("Cannot create Flow without name, and no default_name specified")
                name = self.default_name

            path = folder / name

        self.path = path
        if not path.exists():
            path.mkdir(parents=True, exist_ok=True)

        if reset:
            self.reset()

        if not self._get_info_path().exists():
            self._store_info()

    @classmethod
    def create(cls, path, **kwargs):
        if isinstance(path, str):
            path = pathlib.Path(path)
        self = cls(path=path)

        for key, value in kwargs.items():
            setattr(self, key, value)
        return self

    def _store_info(self):
        info = {"module": self.__class__.__module__, "class": self.__class__.__name__}

        json.dump(info, self._get_info_path().open("w"))

    def _get_info_path(self):
        return self.path / ".flow"

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.path}")'

    @classmethod
    def from_path(cls, path):
        """
        Load a previously created flow from disk
        """
        info = json.load(open(path / ".flow"))

        # load class
        module = importlib.import_module(info["module"])
        cls = getattr(module, info["class"])

        return cls(path=path)

    def reset(self):
        """
        Remove all files in this flow
        """
        shutil.rmtree(self.path)
        self.path.mkdir(parents=True, exist_ok=True)

    def __getstate__(self):
        raise TypeError("This class cannot be pickled.")

    def get(self, k):
        return self.__class__.__dict__[k]


class Linked(Obj):
    """
    A link to another flow on disk
    """

    def __get__(self, obj, type=None):
        if obj is not None:
            name = "_" + self.name
            if not hasattr(obj, name):
                path = obj.path / self.name

                if not path.exists():
                    raise FileNotFoundError(f"File {path} does not exist")
                if not path.is_symlink():
                    raise FileNotFoundError(f"File {path} is not a symlink")

                value = Flow.from_path(path.resolve())
                setattr(obj, name, value)

            return getattr(obj, name)

    def __set__(self, obj, value):
        # symlink to value.path
        path = obj.path / self.name
        name = "_" + self.name
        if path.exists():
            if not path.is_symlink():
                raise FileExistsError(f"File {path} already exists")
            else:
                path.unlink()
        if not str(value.path).startswith("memory"):
            print(value.path.name)
            path.symlink_to(value.path.resolve())
        setattr(obj, name, value)


class Stored(Obj):
    """
    A python object that is stored on disk using pickle
    """

    def __init__(self, default=None, name=None):
        self.default = default
        self.name = name

    def get_path(self, folder):
        return folder / (self.name + ".pkl")

    def __get__(self, obj, type=None):
        if obj is not None:
            if self.name is None:
                print(obj)
                raise ValueError(obj)
            name = "_" + self.name
            if not hasattr(obj, name):
                path = self.get_path(obj.path)
                if not path.exists():
                    if self.default is None:
                        raise FileNotFoundError(f"File {path} does not exist")
                    else:
                        value = self.default()
                        pickle.dump(value, path.open("wb"))
                setattr(obj, name, pickle.load(self.get_path(obj.path).open("rb")))
            return getattr(obj, name)

    def __set__(self, obj, value):
        name = "_" + self.name
        pickle.dump(value, self.get_path(obj.path).open("wb"))
        setattr(obj, name, value)

    def exists(self, obj):
        return self.get_path(obj.path).exists()


class StoredDataFrame(Stored):
    """
    A pandas dataframe stored on disk
    """

    def __init__(self, index_name=None, name=None):
        super().__init__(name=name)
        self.index_name = index_name

    def __set__(self, obj, value):
        if self.index_name is not None:
            value.index.name = self.index_name
        super().__set__(obj, value)


class StoredTensor(Stored):
    def __init__(self, dtype=None, name=None):
        super().__init__(name=name)
        self.dtype = dtype

    def __get__(self, obj, type=None):
        if obj is not None:
            name = "_" + self.name
            if not hasattr(obj, name):
                x = pickle.load(self.get_path(obj.path).open("rb"))
                if not torch.is_tensor(x):
                    raise ValueError(f"File {self.get_path(obj.path)} is not a tensor")
                elif x.dtype is not self.dtype:
                    x = x.to(self.dtype)
                if not x.is_contiguous():
                    x = x.contiguous()
                setattr(obj, name, x)
            return getattr(obj, name)

    def __set__(self, obj, value):
        if not torch.is_tensor(value):
            raise ValueError("Value is not a tensor")
        elif self.dtype is not None:
            value = value.to(self.dtype).contiguous()
        name = "_" + self.name
        pickle.dump(value, self.get_path(obj.path).open("wb"))
        setattr(obj, name, value)


class StoredNumpyInt64(Stored):
    """
    A numpy int64 tensor stored on disk
    """

    def __get__(self, obj, type=None):
        if obj is not None:
            name = "_" + self.name
            if not hasattr(obj, name):
                x = pickle.load(self.get_path(obj.path).open("rb"))
                if x.dtype is not np.int64:
                    x = x.astype(np.int64)
                if not x.flags["C_CONTIGUOUS"]:
                    x = np.ascontiguousarray(x)
                setattr(obj, name, x)
            return getattr(obj, name)

    def __set__(self, obj, value):
        value = np.ascontiguousarray(value.astype(np.int64))
        name = "_" + self.name
        pickle.dump(value, self.get_path(obj.path).open("wb"))
        setattr(obj, name, value)


class CompressedNumpy(Stored):
    """
    A compressed numpy array stored on disk
    """

    dtype = np.float64

    def __get__(self, obj, type=None):
        if obj is not None:
            name = "_" + self.name
            if not hasattr(obj, name):
                x = pickle.load(gzip.GzipFile(self.get_path(obj.path), "rb", compresslevel=3))
                if x.dtype is not self.dtype:
                    x = x.astype(self.dtype)
                if not x.flags["C_CONTIGUOUS"]:
                    x = np.ascontiguousarray(x)
                setattr(obj, name, x)
            return getattr(obj, name)

    def __set__(self, obj, value):
        value = np.ascontiguousarray(value.astype(self.dtype))
        name = "_" + self.name
        pickle.dump(value, gzip.GzipFile(self.get_path(obj.path), "wb", compresslevel=3))
        setattr(obj, name, value)


class CompressedNumpyFloat64(CompressedNumpy):
    dtype = np.float64


class CompressedNumpyInt64(CompressedNumpy):
    dtype = np.int64


class TSV(Stored):
    """
    A pandas object stored on disk in tsv format
    """

    def __init__(self, columns=None, index_name=None, name=None):
        super().__init__(name=name)
        self.columns = columns
        self.index_name = index_name

    def get_path(self, folder):
        return folder / (self.name + ".tsv")

    def __get__(self, obj=None, type=None):
        if obj is not None:
            name = "_" + self.name
            if not hasattr(obj, name):
                import pandas as pd

                x = pd.read_table(self.get_path(obj.path), index_col=0)
                setattr(obj, name, x)
            return getattr(obj, name)

    def __set__(self, obj, value, folder=None):
        name = "_" + self.name
        if folder is None:
            folder = obj.path
        if self.index_name is not None:
            value.index.name = self.index_name
        value.to_csv(self.get_path(folder).open("w"), sep="\t")
        setattr(obj, name, value)


class StoredDict(Obj):
    def __init__(self, cls, name=None):
        super().__init__(name=name)
        self.cls = cls

    def get_path(self, folder):
        return folder / self.name

    def __get__(self, obj, type=None):
        if obj is not None:
            name = "_" + self.name
            if not hasattr(obj, name):
                x = StoredDictInstance(self.name, self.get_path(obj.path), self.cls, obj)
                setattr(obj, name, x)
            return getattr(obj, name)


class StoredDictInstance:
    def __init__(self, name, path, cls, obj):
        self.dict = {}
        self.cls = cls
        self.obj = obj
        self.name = name
        self.path = path
        if not self.path.exists():
            self.path.mkdir(parents=True)
        for file in self.path.iterdir():
            if file.is_dir():
                raise ValueError(f"Folder {file} in {self.obj.path} is not allowed")
            # key is file name without extension
            key = file.name.split(".")[0]
            self.dict[key] = self.cls(name=key)

    def __getitem__(self, key):
        return self.dict[key].__get__(self)

    def __setitem__(self, key, value):
        if key not in self.dict:
            self.dict[key] = self.cls(name=key)
        self.dict[key].__set__(self, value)

    def items(self):
        for k in self.dict:
            yield k, self[k]

    def keys(self):
        return self.dict.keys()



import numpy as np
import pandas as pd

from chromatinhd.flow import (
    Flow,
    Stored,
)

from chromatinhd.data.fragments import Fragments


class Folds(Flow):
    """
    Folds of multiple cell and gene combinations
    """

    folds: dict = Stored()
    """The folds"""

    def sample_cells(
        self,
        fragments: Fragments,
        n_folds: int,
        n_repeats: int = 1,
        overwrite: bool = False,
    ):
        """
        Sample cells and genes into folds

        Parameters:
            fragments:
                the fragments
            n_folds:
                the number of folds
            n_repeats:
                the number of repeats
            overwrite:
                whether to overwrite existing folds
        """
        if not overwrite and self.get("folds").exists(self):
            return

        folds = []

        for repeat_ix in range(n_repeats):
            generator = np.random.RandomState(repeat_ix)

            cells_all = generator.permutation(fragments.n_cells)

            cell_bins = np.floor((np.arange(len(cells_all)) / (len(cells_all) / n_folds)))

            for i in range(n_folds):
                cells_train = cells_all[cell_bins != i]
                cells_validation_test = cells_all[cell_bins == i]
                cells_validation = cells_validation_test[: (len(cells_validation_test) // 2)]
                cells_test = cells_validation_test[(len(cells_validation_test) // 2) :]

                folds.append(
                    {
                        "cells_train": cells_train,
                        "cells_validation": cells_validation,
                        "cells_test": cells_test,
                        "repeat": repeat_ix,
                    }
                )
        self.folds = folds

    def sample_cellxgene(
        self,
        fragments: Fragments,
        n_folds: int,
        n_repeats: int = 1,
        overwrite: bool = False,
    ):
        """
        Sample cells and genes into folds

        Parameters:
            fragments:
                the fragments
            n_folds:
                the number of folds
            n_repeats:
                the number of repeats
            overwrite:
                whether to overwrite existing folds
        """
        if not overwrite and self.get("folds").exists(self):
            return

        folds = []

        for repeat_ix in range(n_repeats):
            generator = np.random.RandomState(repeat_ix)

            cells_all = generator.permutation(fragments.n_cells)

            cell_bins = np.floor((np.arange(len(cells_all)) / (len(cells_all) / n_folds)))

            genes_all = np.arange(fragments.n_genes)

            chr_order = generator.permutation(fragments.regions.coordinates["chr"].unique())
            gene_chrs = pd.Categorical(fragments.regions.coordinates["chr"].astype(str), categories=chr_order).codes
            gene_bins = np.floor((gene_chrs / (len(chr_order) / n_folds))).astype(int)

            for i in range(n_folds):
                cells_train = cells_all[cell_bins != i]
                cells_validation_test = cells_all[cell_bins == i]
                cells_validation = cells_validation_test[: (len(cells_validation_test) // 2)]
                cells_test = cells_validation_test[(len(cells_validation_test) // 2) :]

                genes_train = genes_all[gene_bins != i]
                genes_validation_test = genes_all[gene_bins == i]
                genes_validation = genes_validation_test[: (len(genes_validation_test) // 2)]
                genes_test = genes_validation_test[(len(genes_validation_test) // 2) :]

                folds.append(
                    {
                        "cells_train": cells_train,
                        "cells_validation": cells_validation,
                        "cells_test": cells_test,
                        "genes_train": genes_train,
                        "genes_validation": genes_validation,
                        "genes_test": genes_test,
                        "repeat": repeat_ix,
                    }
                )
        self.folds = folds

    def __getitem__(self, ix):
        return self.folds[ix]

    def __len__(self):
        return len(self.folds)



import pandas as pd

from chromatinhd.flow import Flow, Stored, StoredDataFrame


class Clustering(Flow):
    labels = Stored()
    "Labels for each cell."

    cluster_info = StoredDataFrame(index_name="cluster")
    "Dataframe containing information for each cluster, such as a label."

    @classmethod
    def from_labels(cls, labels, path=None):
        clustering = cls(path)
        if not isinstance(labels, pd.Series):
            labels = pd.Series(labels).astype("category")
        elif not labels.dtype.name == "category":
            labels = labels.astype("category")
        clustering.labels = labels
        clustering.cluster_info = (
            pd.DataFrame(
                {
                    "cluster": labels.unique(),
                    "n_cells": labels.value_counts(),
                    "label": labels.unique(),
                }
            )
            .set_index("cluster")
            .loc[labels.cat.categories]
        )
        return clustering

    @property
    def n_clusters(self):
        return len(self.labels.unique())



from __future__ import annotations
from chromatinhd.flow import Flow, Stored, TSV
import pandas as pd
import numpy as np
import pathlib
import tqdm.auto as tqdm
from chromatinhd.flow import PathLike


class Regions(Flow):
    """
    Regions, typically centered around a transcription start site
    """

    coordinates = TSV(columns=["chrom", "start", "end"])
    window = Stored()

    @classmethod
    def from_transcripts(cls, transcripts: pd.DataFrame, window: [list, np.ndarray], path: pathlib.Path) -> Regions:
        """
        Create regions from a Dataframe of canonical transcripts,
        using a specified window around each transcription start site.

        Parameters:
            transcripts:
                Dataframe of canonical transcripts, with columns chrom, start, end, strand, ensembl_transcript_id
            window:
                Window around each transcription start site. Should be a 2-element array, e.g. [-10000, 10000]
            path:
                Folder in which the regions data will be stored
        """
        transcripts["tss"] = transcripts["start"] * (transcripts["strand"] == 1) + transcripts["end"] * (
            transcripts["strand"] == -1
        )

        regions = transcripts[["chrom", "tss", "ensembl_transcript_id"]].copy()

        regions["strand"] = transcripts["strand"]
        regions["positive_strand"] = (regions["strand"] == 1).astype(int)
        regions["negative_strand"] = (regions["strand"] == -1).astype(int)
        regions["chrom"] = transcripts.loc[regions.index, "chrom"]

        regions["start"] = regions["tss"] + window[0] * (regions["strand"] == 1) - window[1] * (regions["strand"] == -1)
        regions["end"] = regions["tss"] + window[1] * (regions["strand"] == -1) - window[0] * (regions["strand"] == 1)

        return cls.create(
            path=path,
            coordinates=regions[["chrom", "start", "end", "tss", "strand", "ensembl_transcript_id"]],
            window=window,
        )

    def filter_genes(self, genes, path=None) -> Regions:
        """
        Filter genes to those in the regions

        Parameters:
            genes:
                Genes to filter. Should be a pandas Series with the index being the ensembl transcript ids.
            path:
                Path to store the filtered regions
        Returns:
            Regions with only the specified genes
        """

        return Regions.create(coordinates=self.coordinates.loc[genes], window=self.window, path=path)


def select_tss_from_fragments(
    transcripts: pd.DataFrame, fragments_file: PathLike, window: [np.ndarray, tuple] = (-100, 100)
) -> pd.DataFrame:
    """
    Select the TSS with the most fragments within a window of the TSS

    Parameters:
        transcripts:
            Dataframe of transcripts, with columns chrom, tss, ensembl_gene_id
        fragments_file:
            Path to fragments file
        window:
            Window around the TSS to count fragments
    Returns:
        Dataframe of transcripts, with columns chrom, tss, ensembl_gene_id, n_fragments
    """
    if not ([col in transcripts.columns for col in ["chrom", "tss", "ensembl_gene_id"]]):
        raise ValueError("Transcripts should have columns chrom, tss, ensembl_gene_id. ")

    import pysam

    fragments_tabix = pysam.TabixFile(str(fragments_file))

    nfrags = []
    for chrom, tss in tqdm.tqdm(zip(transcripts["chrom"], transcripts["tss"]), total=transcripts.shape[0]):
        frags = list(fragments_tabix.fetch(chrom, tss + window[0], tss + window[1]))
        nfrags.append(len(frags))
    transcripts["n_fragments"] = nfrags
    selected_transcripts = (
        transcripts.reset_index().sort_values("n_fragments", ascending=False).groupby("ensembl_gene_id").first()
    )
    selected_transcripts.index.name = "gene"
    return selected_transcripts



import torch


def _get_default_device():
    """Pick GPU if available, else CPU"""
    if torch.cuda.is_available():
        return torch.device("cuda")
    else:
        return torch.device("cpu")


default_device = _get_default_device()


def get_default_device():
    global default_device
    return default_device


def set_default_device(device):
    global default_device
    default_device = device



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.7
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Preprocess

# %%
from IPython import get_ipython

if get_ipython():
    get_ipython().run_line_magic("load_ext", "autoreload")
    get_ipython().run_line_magic("autoreload", "2")

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import matplotlib as mpl

import seaborn as sns

sns.set_style("ticks")

import torch

import pickle

import scanpy as sc

import tqdm.auto as tqdm
import io

# %%
import chromatinhd as chd

# %%
folder_root = chd.get_output()
folder_data = folder_root / "data"

dataset_name = "pbmc10k"
main_url = "https://cf.10xgenomics.com/samples/cell-arc/2.0.0/pbmc_granulocyte_sorted_10k/pbmc_granulocyte_sorted_10k"
genome = "GRCh38"

folder_data_preproc = folder_data / dataset_name
folder_data_preproc.mkdir(exist_ok=True, parents=True)

# %% [markdown]
# ## Download

# %%
# ! mkdir -p {folder_data_preproc}

# %%
import pathlib

# %%
file = folder_data_preproc/"filtered_feature_bc_matrix.h5"
if not file.exists():
    # !wget {main_url}_filtered_feature_bc_matrix.h5 -O {file}

# %%
if pathlib.Path(chd.get_git_root().parent / "chromatinhd_manuscript" / "output/data" / dataset_name / "atac_fragments.tsv.gz").exists():
    # !rm {folder_data_preproc}/fragments.tsv.gz
    # !ln -s {chd.get_git_root().parent / "chromatinhd_manuscript" / "output/data" / dataset_name / "atac_fragments.tsv.gz"} {folder_data_preproc}/fragments.tsv.gz
    # !wget {main_url}_atac_fragments.tsv.gz.tbi -O {folder_data_preproc}/fragments.tsv.gz.tbi

# %% [markdown]
# ## Create transcriptome

# %%
transcriptome = chd.data.Transcriptome(folder_data_preproc / "transcriptome")

# %% [markdown]
# ### Read and process

# %%
adata = sc.read_10x_h5(folder_data_preproc / "filtered_feature_bc_matrix.h5")

# %%
adata.var.index.name = "symbol"
adata.var = adata.var.reset_index()
adata.var.index = adata.var["gene_ids"]
adata.var.index.name = "gene"

# %%
print(adata.obs.shape[0])
sc.pp.filter_cells(adata, min_counts=1000)
print(adata.obs.shape[0])
sc.pp.filter_cells(adata, min_genes=200)
print(adata.obs.shape[0])
sc.pp.filter_genes(adata, min_cells=100)
print(adata.var.shape[0])

# %%
transcripts = chd.biomart.get_transcripts(chd.biomart.Dataset.from_genome(genome), gene_ids=adata.var.index.unique())
pickle.dump(transcripts, (folder_data_preproc / 'transcripts.pkl').open("wb"))

# %%
# only retain genes that have at least one ensembl transcript
adata = adata[:, adata.var.index.isin(transcripts["ensembl_gene_id"])]

# %%
sc.external.pp.scrublet(adata)

# %%
adata.obs["doublet_score"].plot(kind="hist")

# %%
adata.obs["doublet"] = (adata.obs["doublet_score"] > 0.1).astype("category")

print(adata.obs.shape[0])
adata = adata[~adata.obs["doublet"].astype(bool)]
print(adata.obs.shape[0])

# %%
size_factor = np.median(np.array(adata.X.sum(1)))
adata.uns["size_factor"] = size_factor

# %%
adata.raw = adata

# %%
sc.pp.normalize_total(adata, size_factor)

# %%
sc.pp.log1p(adata)

# %%
sc.pp.pca(adata)

# %%
sc.pp.highly_variable_genes(adata)

# %%
sc.pp.neighbors(adata)
sc.tl.umap(adata)

# %%
import pickle

# %%
adata.layers["normalized"] = adata.X
adata.layers["counts"] = adata.raw.X

# %%
import magic

magic_operator = magic.MAGIC(knn=30, solver = "approximate")
X_smoothened = magic_operator.fit_transform(adata.X)
adata.layers["magic"] = X_smoothened

# %%
pickle.dump(adata, (folder_data_preproc / 'adata.pkl').open("wb"))

# %% [markdown]
# ## Interpret and subset

# %%
adata = pickle.load((folder_data_preproc / 'adata.pkl').open("rb"))

# %%
sc.tl.leiden(adata, resolution=2.0)
sc.pl.umap(adata, color="leiden")

# %%
sc.tl.rank_genes_groups(adata, "leiden", method="wilcoxon", key_added  = "wilcoxon")
# genes_oi = sc.get.rank_genes_groups_df(adata, group=None, key='rank_genes_groups').sort_values("scores", ascending = False).groupby("group").head(2)["names"]

# %%
sc.pl.rank_genes_groups_dotplot(adata, n_genes=100, key="wilcoxon", groupby="leiden", gene_symbols="symbol", groups = ["14"])

# %%
import io
marker_annotation = pd.read_table(
    io.StringIO(
        """ix	symbols	celltype
0	IL7R, CD3D	CD4 naive T
0	IL7R, CD3D, ITGB1	CD4 memory T
1	CD14, LYZ	CD14+ Monocytes
2	MS4A1, IL4R, CD79A	naive B
2	MS4A1, CD79A, TNFRSF13B	memory B
3	CD8A, CD3D	CD8 naive T
4	GNLY, NKG7, GZMA, GZMB, NCAM1	NK
4	GNLY, NKG7, CD3D, CCL5, GZMA, CD8A	CD8 activated T
4	SLC4A10	MAIT
5	FCGR3A, MS4A7	FCGR3A+ Monocytes
5	CD27, JCHAIN	Plasma
6	TCF4	pDCs
6	CST3, CD1C	cDCs
"""
    )
).set_index("celltype")
marker_annotation["symbols"] = marker_annotation["symbols"].str.split(", ")


# %%
# Define cluster score for all markers
def evaluate_partition(
    anndata, marker_dict, gene_symbol_key=None, partition_key="louvain_r1"
):
    # Inputs:
    #    anndata         - An AnnData object containing the data set and a partition
    #    marker_dict     - A dictionary with cell-type markers. The markers should be stores as anndata.var_names or
    #                      an anndata.var field with the key given by the gene_symbol_key input
    #    gene_symbol_key - The key for the anndata.var field with gene IDs or names that correspond to the marker
    #                      genes
    #    partition_key   - The key for the anndata.obs field where the cluster IDs are stored. The default is
    #                      'louvain_r1'

    # Test inputs
    if partition_key not in anndata.obs.columns.values:
        print("KeyError: The partition key was not found in the passed AnnData object.")
        print(
            "   Have you done the clustering? If so, please tell pass the cluster IDs with the AnnData object!"
        )
        raise

    if (gene_symbol_key != None) and (
        gene_symbol_key not in anndata.var.columns.values
    ):
        print(
            "KeyError: The provided gene symbol key was not found in the passed AnnData object."
        )
        print(
            "   Check that your cell type markers are given in a format that your anndata object knows!"
        )
        raise

    if gene_symbol_key:
        gene_ids = anndata.var[gene_symbol_key]
    else:
        gene_ids = anndata.var_names

    clusters = np.unique(anndata.obs[partition_key])
    n_clust = len(clusters)
    n_groups = len(marker_dict)

    marker_res = np.zeros((n_groups, n_clust))
    z_scores = sc.pp.scale(anndata, copy=True)

    i = 0
    for group in marker_dict:
        # Find the corresponding columns and get their mean expression in the cluster
        j = 0
        for clust in clusters:
            cluster_cells = np.in1d(z_scores.obs[partition_key], clust)
            marker_genes = np.in1d(gene_ids, marker_dict[group])
            marker_res[i, j] = z_scores.X[np.ix_(cluster_cells, marker_genes)].mean()
            j += 1
        i += 1

    variances = np.nanvar(marker_res, axis=0)
    if np.all(np.isnan(variances)):
        print(
            "No variances could be computed, check if your cell markers are in the data set."
        )
        print(
            "Maybe the cell marker IDs do not correspond to your gene_symbol_key input or the var_names"
        )
        raise

    marker_res_df = pd.DataFrame(marker_res, columns=clusters, index=marker_dict.keys())

    # Return the median of the variances over the clusters
    return marker_res_df



# %%
cluster_celltypes = evaluate_partition(
    adata, marker_annotation["symbols"].to_dict(), "symbol", partition_key="leiden"
).idxmax()

adata.obs["celltype"] = adata.obs["celltype"] = cluster_celltypes[
    adata.obs["leiden"]
].values
adata.obs["celltype"] = adata.obs["celltype"] = adata.obs[
    "celltype"
].astype(str)

# %%
adata.obs["log_n_counts"] = np.log(adata.obs["n_counts"])
sc.pl.umap(adata, color=["celltype", "log_n_counts", "leiden"], legend_loc="on data")

# %%
adata = adata[~adata.obs["leiden"].isin(["21", "23"])]

# %%
sc.pl.umap(adata, color=["celltype", "log_n_counts", "leiden"], legend_loc="on data")

# %%
pickle.dump(adata, (folder_data_preproc / 'adata_annotated.pkl').open("wb"))

# %% [markdown]
# ## TSS

# %%
adata = pickle.load((folder_data_preproc / 'adata_annotated.pkl').open("rb"))

# %%
transcripts = pickle.load((folder_data_preproc / 'transcripts.pkl').open("rb"))
transcripts = transcripts.loc[transcripts["ensembl_gene_id"].isin(adata.var.index)]

# %%
fragments_file = folder_data_preproc / "fragments.tsv.gz"
selected_transcripts = chd.data.regions.select_tss_from_fragments(transcripts, fragments_file)

# %%
plt.scatter(adata.var["means"], np.log(transcripts.groupby("ensembl_gene_id")["nfrags"].max()))

# %%
pickle.dump(selected_transcripts, (folder_data_preproc / 'selected_transcripts.pkl').open("wb"))



from .dataset import Dataset
from .tss import get_canonical_transcripts, get_exons, get_transcripts
from . import tss

__all__ = ["Dataset", "get_canonical_transcripts", "get_exons", "get_transcripts", "tss"]



import numpy as np
import pandas as pd
import pathlib
from typing import Union

from chromatinhd.flow import Flow, Stored, StoredDict, TSV
from chromatinhd import sparse


class Transcriptome(Flow):
    """
    A transcriptome containing counts for each gene in each cell.
    """

    var: pd.DataFrame = TSV(index_name="gene")
    obs: pd.DataFrame = TSV(index_name="cell")

    adata = Stored()
    "Anndata object containing the transcriptome data."

    def gene_id(self, symbol, column="symbol"):
        """
        Get the gene id for a given gene symbol.
        """
        assert all(pd.Series(symbol).isin(self.var[column])), set(
            pd.Series(symbol)[~pd.Series(symbol).isin(self.var[column])]
        )
        return self.var.reset_index("gene").set_index(column).loc[symbol]["gene"]

    def symbol(self, gene_id, column="symbol"):
        """
        Get the gene symbol for a given gene ID (e.g. Ensembl ID).
        """
        assert all(pd.Series(gene_id).isin(self.var.index)), set(
            pd.Series(gene_id)[~pd.Series(gene_id).isin(self.var.index)]
        )
        return self.var.loc[gene_id][column]

    def gene_ix(self, symbol):
        """
        Get the gene index for a given gene symbol.
        """
        self.var["ix"] = np.arange(self.var.shape[0])
        assert all(pd.Series(symbol).isin(self.var["symbol"])), set(
            pd.Series(symbol)[~pd.Series(symbol).isin(self.var["symbol"])]
        )
        return self.var.reset_index("gene").set_index("symbol").loc[symbol]["ix"]

    def create_X(self):
        X_scipy = self.adata.X
        if isinstance(X_scipy, np.ndarray):
            import scipy.sparse

            X_scipy = scipy.sparse.csr_matrix(X_scipy)
        X = sparse.COOMatrix.from_scipy_csr(X_scipy)
        X.populate_mapping()

        self.X = X

    X = Stored()
    "The main transcriptome data, typically normalized counts."

    @classmethod
    def from_adata(cls, adata, path: Union[pathlib.Path, str] = None):
        """
        Create a Transcriptome object from an AnnData object.

        Parameters:
            adata:
                Anndata object containing the transcriptome data.
            path:
                Folder in which the transcriptome data will be stored.
        """
        transcriptome = cls(path=path)
        transcriptome.adata = adata

        for k, v in adata.layers.items():
            transcriptome.layers[k] = v
        transcriptome.X = adata.X
        transcriptome.var = adata.var
        transcriptome.obs = adata.obs
        return transcriptome

    layers = StoredDict(Stored)
    "Dictionary of layers, such as raw, normalized and imputed data."

    def filter_genes(self, genes, path=None):
        """
        Filter genes

        Parameters:
            genes:
                Genes to filter. Should be a pandas Series with the index being the ensembl transcript ids.
        """

        self.var["ix"] = np.arange(self.var.shape[0])
        gene_ixs = self.var["ix"].loc[genes]

        layers = {}
        for k, v in self.layers.items():
            layers[k] = v[:, gene_ixs]
        X = self.X[:, gene_ixs]

        return Transcriptome.create(
            var=self.var.loc[genes],
            obs=self.obs,
            X=X,
            layers=layers,
            path=path,
        )

    def get_X(self, gene_ids):
        """
        Get the counts for a given set of genes.
        """
        gene_ixs = self.var.index.get_loc(gene_ids)
        value = self.X[:, gene_ixs]

        if sparse.is_scipysparse(value):
            value = np.array(value.todense())
            if isinstance(gene_ids, str):
                value = value[:, 0]
        return value


class ClusterTranscriptome(Flow):
    var = Stored()
    obs = Stored()
    adata = Stored()
    X = Stored()

    def gene_id(self, symbol):
        assert all(pd.Series(symbol).isin(self.var["symbol"])), set(
            pd.Series(symbol)[~pd.Series(symbol).isin(self.var["symbol"])]
        )
        return self.var.reset_index("gene").set_index("symbol").loc[symbol]["gene"]

    def symbol(self, gene_id):
        assert all(pd.Series(gene_id).isin(self.var.index)), set(
            pd.Series(gene_id)[~pd.Series(gene_id).isin(self.var.index)]
        )
        return self.var.loc[gene_id]["symbol"]

    def gene_ix(self, symbol):
        self.var["ix"] = np.arange(self.var.shape[0])
        assert all(pd.Series(symbol).isin(self.var["symbol"])), set(
            pd.Series(symbol)[~pd.Series(symbol).isin(self.var["symbol"])]
        )
        return self.var.reset_index("gene").set_index("symbol").loc[symbol]["ix"]


class ClusteredTranscriptome(Flow):
    donors_info = Stored()
    clusters_info = Stored()
    var = Stored()
    X = Stored()

    def gene_id(self, symbol):
        assert all(pd.Series(symbol).isin(self.var["symbol"])), set(
            pd.Series(symbol)[~pd.Series(symbol).isin(self.var["symbol"])]
        )
        return self.var.reset_index("gene").set_index("symbol").loc[symbol]["gene"]

    def symbol(self, gene_id):
        assert all(pd.Series(gene_id).isin(self.var.index)), set(
            pd.Series(gene_id)[~pd.Series(gene_id).isin(self.var.index)]
        )
        return self.var.loc[gene_id]["symbol"]

    def gene_ix(self, symbol):
        self.var["ix"] = np.arange(self.var.shape[0])
        assert all(pd.Series(symbol).isin(self.var["symbol"])), set(
            pd.Series(symbol)[~pd.Series(symbol).isin(self.var["symbol"])]
        )
        return self.var.reset_index("gene").set_index("symbol").loc[symbol]["ix"]



import pandas as pd
from .dataset import Dataset


def get_transcripts(
    biomart_dataset: Dataset,
    gene_ids=None,
    chrom=None,
    start=None,
    end=None,
    filter_chromosomes=True,
    filter_protein_coding=True,
) -> pd.DataFrame:
    """
    Get all canonical transcripts
    """
    batch_size = 100
    transcripts_ = []

    for i in range(0, len(gene_ids), batch_size):
        filters = []
        if gene_ids is not None:
            filters.append(biomart_dataset.filter("ensembl_gene_id", value=",".join(gene_ids[i : i + batch_size])))
        if chrom is not None:
            filters.append(biomart_dataset.filter("chromosome_name", value=str(chrom).replace("chr", "")))
        if start is not None:
            filters.append(biomart_dataset.filter("start", value=str(start)))
        if end is not None:
            filters.append(biomart_dataset.filter("end", value=str(end)))
        transcripts_.append(
            biomart_dataset.get(
                [
                    biomart_dataset.attribute("ensembl_gene_id"),
                    biomart_dataset.attribute("transcript_start"),
                    biomart_dataset.attribute("transcript_end"),
                    biomart_dataset.attribute("end_position"),
                    biomart_dataset.attribute("start_position"),
                    biomart_dataset.attribute("ensembl_transcript_id"),
                    biomart_dataset.attribute("chromosome_name"),
                    biomart_dataset.attribute("strand"),
                    biomart_dataset.attribute("external_gene_name"),
                    biomart_dataset.attribute("transcript_is_canonical"),
                    biomart_dataset.attribute("transcript_biotype"),
                ],
                filters=filters,
            )
        )
    transcripts = pd.concat(transcripts_)
    transcripts["chrom"] = "chr" + transcripts["chromosome_name"].astype(str)
    transcripts = transcripts.set_index("ensembl_transcript_id")
    transcripts = transcripts.rename(
        columns={
            "transcript_start": "start",
            "transcript_end": "end",
            "external_gene_name": "symbol",
        }
    )
    transcripts["tss"] = transcripts["start"] * (transcripts["strand"] == 1) + transcripts["end"] * (
        transcripts["strand"] == -1
    )

    transcripts = transcripts.drop(
        [
            "chromosome_name",
            "start_position",
            "end_position",
        ],
        axis=1,
    )

    # filter on protein coding
    if filter_protein_coding:
        transcripts = transcripts.loc[
            transcripts["transcript_biotype"].isin(["protein_coding", "protein_coding_CDS_not_defined"])
        ]

    if filter_chromosomes:
        transcripts = transcripts.loc[~transcripts["chrom"].isin(["chrM", "chrMT"])]
        transcripts = transcripts.loc[~transcripts["chrom"].str.contains("_")]
        transcripts = transcripts.loc[~transcripts["chrom"].str.contains("\.")]

    return transcripts


def get_canonical_transcripts(
    biomart_dataset: Dataset,
    gene_ids=None,
    chrom=None,
    start=None,
    end=None,
    filter_chromosomes: bool = True,
    filter_protein_coding: bool = True,
) -> pd.DataFrame:
    """
    Get all canonical transcripts
    """
    filters = []
    if gene_ids is not None:
        filters.append(biomart_dataset.filter("ensembl_gene_id", value=",".join(gene_ids)))
    if chrom is not None:
        filters.append(biomart_dataset.filter("chromosome_name", value=str(chrom).replace("chr", "")))
    if start is not None:
        filters.append(biomart_dataset.filter("start", value=str(start)))
    if end is not None:
        filters.append(biomart_dataset.filter("end", value=str(end)))
    genes = biomart_dataset.get(
        [
            biomart_dataset.attribute("ensembl_gene_id"),
            biomart_dataset.attribute("transcript_start"),
            biomart_dataset.attribute("transcript_end"),
            biomart_dataset.attribute("end_position"),
            biomart_dataset.attribute("start_position"),
            biomart_dataset.attribute("ensembl_transcript_id"),
            biomart_dataset.attribute("chromosome_name"),
            biomart_dataset.attribute("strand"),
            biomart_dataset.attribute("external_gene_name"),
            biomart_dataset.attribute("transcript_is_canonical"),
            biomart_dataset.attribute("transcript_biotype"),
        ],
        filters=filters,
    )
    genes["chrom"] = "chr" + genes["chromosome_name"].astype(str)
    genes = genes.sort_values("transcript_is_canonical").groupby("ensembl_gene_id").first()
    genes = genes.rename(
        columns={
            "transcript_start": "start",
            "transcript_end": "end",
            "external_gene_name": "symbol",
        }
    )
    genes.index.name = "gene"
    genes = genes.drop(
        [
            "chromosome_name",
            "start_position",
            "end_position",
        ],
        axis=1,
    )

    # add tss
    genes["tss"] = genes["start"] * (genes["strand"] == 1) + genes["end"] * (genes["strand"] == -1)

    # remove genes without a symbol
    genes = genes.loc[~pd.isnull(genes["symbol"])]

    if filter_chromosomes:
        genes = genes.loc[~genes["chrom"].str.contains("_")]
        genes = genes.loc[~genes["chrom"].isin(["chrMT", "chrM"])]

    # filter on protein coding
    if filter_protein_coding:
        genes = genes.loc[genes["transcript_biotype"].isin(["protein_coding", "protein_coding_CDS_not_defined"])]

    # order according to gene_ids
    genes = genes.reindex(gene_ids)

    # add ensembl gene id
    genes["ensembl_gene_id"] = genes.index

    return genes


def get_exons(biomart_dataset: Dataset, chrom, start, end):
    canonical_transcripts = get_canonical_transcripts(biomart_dataset, chrom=chrom, start=start, end=end)
    exons = biomart_dataset.get(
        [
            biomart_dataset.attribute("ensembl_gene_id"),
            biomart_dataset.attribute("exon_chrom_start"),
            biomart_dataset.attribute("exon_chrom_end"),
            biomart_dataset.attribute("genomic_coding_start"),
            biomart_dataset.attribute("genomic_coding_end"),
            biomart_dataset.attribute("ensembl_transcript_id"),
        ],
        filters=[
            biomart_dataset.filter("start", value=str(start)),
            biomart_dataset.filter("end", value=str(end)),
            biomart_dataset.filter("chromosome_name", value=str(chrom).replace("chr", "")),
            biomart_dataset.filter(
                "ensembl_transcript_id",
                value=",".join(canonical_transcripts["ensembl_transcript_id"]),
            ),
        ],
    )

    return exons



"""
Additive model for predicting gene expression from fragments
"""

import pickle

import numpy as np
import pandas as pd
import torch
import torch_scatter
import xarray as xr

from chromatinhd.flow import Flow, Stored
from chromatinhd.loaders import LoaderPool
from chromatinhd.models import HybridModel
from chromatinhd.models.pred.loader.minibatches import Minibatcher
from chromatinhd.models.pred.loader.transcriptome_fragments import (
    TranscriptomeFragments,
)
from chromatinhd.models.pred.trainer import Trainer2
from chromatinhd.optim import SparseDenseAdam

from chromatinhd import get_default_device

from .loss import gene_paircor_loss, paircor, paircor_loss


class SineEncoding(torch.nn.Module):
    def __init__(self, n_frequencies, add_length=False):
        super().__init__()

        self.register_buffer(
            "frequencies",
            torch.tensor([[1 / 1000 ** (2 * i / n_frequencies)] * 2 for i in range(1, n_frequencies + 1)]).flatten(-2),
        )
        self.register_buffer(
            "shifts",
            torch.tensor([[0, torch.pi / 2] for _ in range(1, n_frequencies + 1)]).flatten(-2),
        )

        self.n_embedding_dimensions = n_frequencies * 2 * 2

        if add_length:
            self.n_embedding_dimensions += n_frequencies * 2
        self.add_length = add_length

    def forward(self, coordinates):
        embedding = torch.sin((coordinates[..., None] * self.frequencies + self.shifts).flatten(-2))
        if self.add_length:
            embedding = torch.cat(
                [
                    embedding,
                    torch.sin(
                        (
                            (coordinates[..., 0, None] - coordinates[..., 1, None])[..., None] * self.frequencies
                            + self.shifts
                        ).flatten(-2)
                    ),
                ],
                -1,
            )
        return embedding


class FragmentEmbedder(torch.nn.Module):
    dropout_rate = 0.0

    def __init__(
        self,
        n_frequencies=10,
        n_embedding_dimensions=5,
        n_layers=1,
        nonlinear=True,
        dropout_rate=0.0,
        add_length=False,
        add_residual_count=False,
        **kwargs,
    ):
        self.n_embedding_dimensions = n_embedding_dimensions

        self.nonlinear = nonlinear
        self.dropout_rate = dropout_rate

        super().__init__(**kwargs)

        self.sine_encoding = SineEncoding(n_frequencies=n_frequencies, add_length=add_length)

        layers = []
        for layer_ix in range(n_layers):
            if layer_ix == 0:
                layers.append(
                    torch.nn.Linear(
                        self.sine_encoding.n_embedding_dimensions,
                        self.n_embedding_dimensions,
                    )
                )
            else:
                layers.append(torch.nn.Linear(self.n_embedding_dimensions, self.n_embedding_dimensions))
            layers.append(torch.nn.Sigmoid())
            if self.dropout_rate > 0:
                layers.append(torch.nn.Dropout(self.dropout_rate))
        layers.append(torch.nn.Linear(self.n_embedding_dimensions, self.n_embedding_dimensions))

        self.nn = torch.nn.Sequential(*layers)

        if add_residual_count:
            self.n_embedding_dimensions += 1
        self.add_residual_count = add_residual_count

    def forward(self, coordinates):
        embedding = self.sine_encoding(coordinates)
        embedding = self.nn(embedding)

        if self.add_residual_count:
            embedding = torch.cat(
                [
                    embedding,
                    torch.ones(
                        (*coordinates.shape[:-1], 1),
                        device=coordinates.device,
                        dtype=torch.float,
                    ),
                ],
                -1,
            )

        return embedding


class FragmentEmbedderCounter(torch.nn.Module):
    """
    Dummy embedding of fragments in a single embedding dimension of ones
    """

    def __init__(self, *args, **kwargs):
        self.n_embedding_dimensions = 1
        super().__init__(*args, **kwargs)

    def forward(self, coordinates):
        return torch.ones((*coordinates.shape[:-1], 1), device=coordinates.device, dtype=torch.float)


class EmbeddingGenePooler(torch.nn.Module):
    """
    Pools fragments across genes and cells
    """

    def __init__(self, reduce="sum"):
        self.reduce = reduce
        super().__init__()

    def forward(self, embedding, fragment_cellxgene_ix, cell_n, gene_n):
        if self.reduce == "mean":
            cellxgene_embedding = torch_scatter.segment_mean_coo(
                embedding, fragment_cellxgene_ix, dim_size=cell_n * gene_n
            )
        elif self.reduce == "sum":
            cellxgene_embedding = torch_scatter.segment_sum_coo(
                embedding, fragment_cellxgene_ix, dim_size=cell_n * gene_n
            )
        else:
            raise ValueError()
        cell_gene_embedding = cellxgene_embedding.reshape((cell_n, gene_n, cellxgene_embedding.shape[-1]))
        return cell_gene_embedding


class EmbeddingToExpression(torch.nn.Module):
    """
    Predicts gene expression using a [cell, gene, component] embedding in a gene-specific manner
    """

    def __init__(self, n_embedding_dimensions=5, n_layers=1):
        self.n_embedding_dimensions = n_embedding_dimensions

        super().__init__()

        layers = []
        for layer_ix in range(n_layers):
            if layer_ix == 0:
                layers.append(torch.nn.Linear(self.n_embedding_dimensions, self.n_embedding_dimensions))
            else:
                layers.append(torch.nn.Linear(self.n_embedding_dimensions, self.n_embedding_dimensions))
            layers.append(torch.nn.Sigmoid())
        layers.append(torch.nn.Linear(self.n_embedding_dimensions, 1))
        self.nn = torch.nn.Sequential(*layers)

    def forward(self, cell_gene_embedding):
        out = self.nn(cell_gene_embedding).squeeze(-1)
        return out


class Model(torch.nn.Module, HybridModel):
    trace = None

    def __init__(
        self,
        dummy=False,
        n_frequencies=50,
        n_fragment_embedder_layers=1,
        n_embedding_to_expression_layers=1,
        reduce="sum",
        nonlinear=True,
        n_embedding_dimensions=10,
        dropout_rate=0.0,
        add_length=False,
        add_residual_count=False,
    ):
        super().__init__()

        if dummy:
            self.fragment_embedder = FragmentEmbedderCounter()
        else:
            self.fragment_embedder = FragmentEmbedder(
                n_frequencies=n_frequencies,
                nonlinear=nonlinear,
                n_embedding_dimensions=n_embedding_dimensions,
                dropout_rate=dropout_rate,
                n_layers=n_fragment_embedder_layers,
                add_length=add_length,
                add_residual_count=add_residual_count,
            )
        self.embedding_gene_pooler = EmbeddingGenePooler(reduce=reduce)
        self.embedding_to_expression = EmbeddingToExpression(
            n_embedding_dimensions=self.fragment_embedder.n_embedding_dimensions,
            n_layers=n_embedding_to_expression_layers if not dummy else 0,
        )

    def forward(self, data):
        fragment_embedding = self.fragment_embedder(data.fragments.coordinates)
        cell_gene_embedding = self.embedding_gene_pooler(
            fragment_embedding,
            data.fragments.local_cellxgene_ix,
            data.minibatch.n_cells,
            data.minibatch.n_genes,
        )
        expression_predicted = self.embedding_to_expression(cell_gene_embedding)
        return expression_predicted

    def forward_loss(self, data):
        expression_predicted = self.forward(data)
        expression_true = data.transcriptome.value
        return paircor_loss(expression_predicted, expression_true)

    def forward_gene_loss(self, data):
        expression_predicted = self.forward(data)
        expression_true = data.transcriptome.value
        return gene_paircor_loss(expression_predicted, expression_true)

    def forward_multiple(self, data, fragments_oi, min_fragments=1):
        fragment_embedding = self.fragment_embedder(data.fragments.coordinates, data.fragments.genemapping)

        total_n_fragments = torch.bincount(
            data.fragments.local_cellxgene_ix,
            minlength=data.minibatch.n_genes * data.minibatch.n_cells,
        ).reshape((data.minibatch.n_cells, data.minibatch.n_genes))

        total_cell_gene_embedding = self.embedding_gene_pooler.forward(
            fragment_embedding,
            data.fragments.local_cellxgene_ix,
            data.minibatch.n_cells,
            data.minibatch.n_genes,
        )

        total_expression_predicted = self.embedding_to_expression.forward(total_cell_gene_embedding)

        for fragments_oi_ in fragments_oi:
            if (fragments_oi_ is not None) and ((~fragments_oi_).sum() > min_fragments):
                lost_fragments_oi = ~fragments_oi_
                lost_local_cellxgene_ix = data.fragments.local_cellxgene_ix[lost_fragments_oi]
                n_fragments = total_n_fragments - torch.bincount(
                    lost_local_cellxgene_ix,
                    minlength=data.minibatch.n_genes * data.minibatch.n_cells,
                ).reshape((data.minibatch.n_cells, data.minibatch.n_genes))
                cell_gene_embedding = total_cell_gene_embedding - self.embedding_gene_pooler.forward(
                    fragment_embedding[lost_fragments_oi],
                    lost_local_cellxgene_ix,
                    data.minibatch.n_cells,
                    data.minibatch.n_genes,
                )

                expression_predicted = self.embedding_to_expression.forward(cell_gene_embedding)
            else:
                n_fragments = total_n_fragments
                expression_predicted = total_expression_predicted

            yield expression_predicted, n_fragments

    def train_model(self, fragments, transcriptome, fold, device=None):
        # set up minibatchers and loaders
        minibatcher_train = Minibatcher(
            fold["cells_train"],
            fold["genes_train"],
            n_genes_step=500,
            n_cells_step=500,
            permute_cells=True,
            permute_genes=True,
        )
        minibatcher_validation = Minibatcher(
            fold["cells_validation"],
            fold["genes_validation"],
            n_genes_step=500,
            n_cells_step=500,
            permute_cells=False,
            permute_genes=False,
        )

        if device is None:
            device = get_default_device()

        loaders_train = LoaderPool(
            TranscriptomeFragments,
            dict(
                transcriptome=transcriptome,
                fragments=fragments,
                cellxgene_batch_size=minibatcher_train.cellxgene_batch_size,
            ),
            n_workers=10,
        )
        loaders_validation = LoaderPool(
            TranscriptomeFragments,
            dict(
                transcriptome=transcriptome,
                fragments=fragments,
                cellxgene_batch_size=minibatcher_validation.cellxgene_batch_size,
            ),
            n_workers=5,
        )

        trainer = Trainer2(
            self,
            loaders_train,
            loaders_validation,
            minibatcher_train,
            minibatcher_validation,
            SparseDenseAdam(
                self.parameters_sparse(),
                self.parameters_dense(),
                lr=1e-3,
                weight_decay=1e-5,
            ),
            n_epochs=30,
            checkpoint_every_epoch=1,
            optimize_every_step=1,
            device=device,
        )

        trainer.train()
        self.trace = trainer.trace
        # trainer.trace.plot()

    def get_prediction(
        self,
        fragments,
        transcriptome,
        cells=None,
        cell_ixs=None,
        genes=None,
        gene_ixs=None,
        device=None,
        return_raw=False,
    ):
        """
        Returns the prediction of a dataset
        """
        if cell_ixs is None:
            if cells is None:
                cells = fragments.obs.index
            fragments.obs["ix"] = np.arange(len(fragments.obs))
            cell_ixs = fragments.obs.loc[cells]["ix"].values
        if cells is None:
            cells = fragments.obs.index[cell_ixs]

        if gene_ixs is None:
            if genes is None:
                genes = fragments.var.index
            fragments.var["ix"] = np.arange(len(fragments.var))
            gene_ixs = fragments.var.loc[genes]["ix"].values
        if genes is None:
            genes = fragments.var.index[gene_ixs]

        if device is None:
            device = get_default_device()

        minibatches = Minibatcher(
            cell_ixs,
            gene_ixs,
            n_genes_step=500,
            n_cells_step=200,
            use_all_cells=True,
            use_all_genes=True,
            permute_cells=False,
            permute_genes=False,
        )
        loaders = LoaderPool(
            TranscriptomeFragments,
            dict(
                transcriptome=transcriptome,
                fragments=fragments,
                cellxgene_batch_size=minibatches.cellxgene_batch_size,
            ),
            n_workers=5,
        )
        loaders.initialize(minibatches)

        predicted = np.zeros((len(cell_ixs), len(gene_ixs)))
        expected = np.zeros((len(cell_ixs), len(gene_ixs)))
        n_fragments = np.zeros((len(cell_ixs), len(gene_ixs)))

        cell_mapping = np.zeros(fragments.n_cells, dtype=np.int64)
        cell_mapping[cell_ixs] = np.arange(len(cell_ixs))

        gene_mapping = np.zeros(fragments.n_genes, dtype=np.int64)
        gene_mapping[gene_ixs] = np.arange(len(gene_ixs))

        self.eval()
        self = self.to(device)

        for data in loaders:
            data = data.to(device)
            with torch.no_grad():
                pred_mb = self.forward(data)
            predicted[
                np.ix_(
                    cell_mapping[data.minibatch.cells_oi],
                    gene_mapping[data.minibatch.genes_oi],
                )
            ] = pred_mb.cpu().numpy()
            expected[
                np.ix_(
                    cell_mapping[data.minibatch.cells_oi],
                    gene_mapping[data.minibatch.genes_oi],
                )
            ] = data.transcriptome.value.cpu().numpy()
            n_fragments[
                np.ix_(
                    cell_mapping[data.minibatch.cells_oi],
                    gene_mapping[data.minibatch.genes_oi],
                )
            ] = (
                torch.bincount(
                    data.fragments.local_cellxgene_ix,
                    minlength=len(data.minibatch.cells_oi) * len(data.minibatch.genes_oi),
                )
                .reshape(len(data.minibatch.cells_oi), len(data.minibatch.genes_oi))
                .cpu()
                .numpy()
            )

        self = self.to("cpu")

        if return_raw:
            return predicted, expected, n_fragments

        result = xr.Dataset(
            {
                "predicted": xr.DataArray(
                    predicted,
                    dims=("cell", "gene"),
                    coords={"cell": cells, "gene": genes},
                ),
                "expected": xr.DataArray(
                    expected,
                    dims=("cell", "gene"),
                    coords={"cell": cells, "gene": genes},
                ),
                "n_fragments": xr.DataArray(
                    n_fragments,
                    dims=("cell", "gene"),
                    coords={"cell": cells, "gene": genes},
                ),
            }
        )
        return result

    def get_prediction_censored(
        self,
        fragments,
        transcriptome,
        censorer,
        cells=None,
        cell_ixs=None,
        genes=None,
        gene_ixs=None,
        device=None,
    ):
        """
        Returns the prediction of multiple censored dataset
        """
        if cell_ixs is None:
            if cells is None:
                cells = fragments.obs.index
            fragments.obs["ix"] = np.arange(len(fragments.obs))
            cell_ixs = fragments.obs.loc[cells]["ix"].values
        if cells is None:
            cells = fragments.obs.index[cell_ixs]

        if gene_ixs is None:
            if genes is None:
                genes = fragments.var.index
            fragments.var["ix"] = np.arange(len(fragments.var))
            gene_ixs = fragments.var.loc[genes]["ix"].values
        if genes is None:
            genes = fragments.var.index[gene_ixs]

        if device is None:
            device = get_default_device()

        minibatcher = Minibatcher(
            cell_ixs,
            gene_ixs,
            n_genes_step=500,
            n_cells_step=5000,
            use_all_cells=True,
            use_all_genes=True,
            permute_cells=False,
            permute_genes=False,
        )
        loaders = LoaderPool(
            TranscriptomeFragments,
            dict(
                transcriptome=transcriptome,
                fragments=fragments,
                cellxgene_batch_size=minibatcher.cellxgene_batch_size,
            ),
            n_workers=10,
        )
        loaders.initialize(minibatcher)

        predicted = np.zeros((len(censorer), len(cell_ixs), len(gene_ixs)), dtype=float)
        expected = np.zeros((len(cell_ixs), len(gene_ixs)), dtype=float)
        n_fragments = np.zeros((len(censorer), len(cell_ixs), len(gene_ixs)), dtype=int)

        cell_mapping = np.zeros(fragments.n_cells, dtype=np.int64)
        cell_mapping[cell_ixs] = np.arange(len(cell_ixs))
        gene_mapping = np.zeros(fragments.n_genes, dtype=np.int64)
        gene_mapping[gene_ixs] = np.arange(len(gene_ixs))

        self.eval()
        self.to(device)
        for data in loaders:
            data = data.to(device)
            fragments_oi = censorer(data)

            with torch.no_grad():
                for design_ix, (
                    pred_mb,
                    n_fragments_oi_mb,
                ) in enumerate(self.forward_multiple(data, fragments_oi)):
                    ix = np.ix_(
                        [design_ix],
                        cell_mapping[data.minibatch.cells_oi],
                        gene_mapping[data.minibatch.genes_oi],
                    )
                    predicted[ix] = pred_mb.cpu().numpy()
                    n_fragments[ix] = n_fragments_oi_mb.cpu().numpy()
            expected[
                np.ix_(
                    cell_mapping[data.minibatch.cells_oi],
                    gene_mapping[data.minibatch.genes_oi],
                )
            ] = data.transcriptome.value.cpu().numpy()

        self.to("cpu")

        return predicted, expected, n_fragments


class Models(Flow):
    n_models = Stored()

    @property
    def models_path(self):
        path = self.path / "models"
        path.mkdir(exist_ok=True)
        return path

    def train_models(self, fragments, transcriptome, folds, device=None):
        self.n_models = len(folds)
        for fold_ix, fold in [(fold_ix, fold) for fold_ix, fold in enumerate(folds)]:
            desired_outputs = [self.models_path / ("model_" + str(fold_ix) + ".pkl")]
            force = False
            if not all([desired_output.exists() for desired_output in desired_outputs]):
                force = True

            if force:
                model = Model()
                model.train_model(fragments, transcriptome, fold, device=device)

                model = model.to("cpu")

                pickle.dump(
                    model,
                    open(self.models_path / ("model_" + str(fold_ix) + ".pkl"), "wb"),
                )

    def __getitem__(self, ix):
        return pickle.load((self.models_path / ("model_" + str(ix) + ".pkl")).open("rb"))

    def __len__(self):
        return self.n_models

    def __iter__(self):
        for ix in range(len(self)):
            yield self[ix]

    def get_gene_cors(self, fragments, transcriptome, folds, device=None):
        cor_predicted = np.zeros((len(fragments.var.index), len(folds)))
        cor_n_fragments = np.zeros((len(fragments.var.index), len(folds)))
        n_fragments = np.zeros((len(fragments.var.index), len(folds)))

        if device is None:
            device = get_default_device()

        for model_ix, (model, fold) in enumerate(zip(self, folds)):
            prediction = model.get_prediction(fragments, transcriptome, cell_ixs=fold["cells_test"], device=device)

            cor_predicted[:, model_ix] = paircor(prediction["predicted"].values, prediction["expected"].values)
            cor_n_fragments[:, model_ix] = paircor(prediction["n_fragments"].values, prediction["expected"].values)

            n_fragments[:, model_ix] = prediction["n_fragments"].values.sum(0)
        cor_predicted = pd.Series(cor_predicted.mean(1), index=fragments.var.index, name="cor_predicted")
        cor_n_fragments = pd.Series(cor_n_fragments.mean(1), index=fragments.var.index, name="cor_n_fragments")
        n_fragments = pd.Series(n_fragments.mean(1), index=fragments.var.index, name="n_fragments")
        result = pd.concat([cor_predicted, cor_n_fragments, n_fragments], axis=1)
        result["deltacor"] = result["cor_predicted"] - result["cor_n_fragments"]

        return result



import chromatinhd as chd


class TestGetCanonicalTranscripts:
    def test_simple(self, dataset_grch38):
        pass
        # transcripts = chd.biomart.tss.get_canonical_transcripts(biomart_dataset = dataset_grch38)

        # assert transcripts.shape[0] > 1000


class TestGetExons:
    def test_simple(self, dataset_grch38):
        pass
        # exons = chd.biomart.tss.get_exons(biomart_dataset = dataset_grch38, chrom = "chr1", start = 1000, end = 2000)



import itertools
import pickle
from typing import List, Optional

import numpy as np
import pandas as pd
import tqdm.auto as tqdm
import xarray as xr

import chromatinhd as chd
from chromatinhd import get_default_device
from chromatinhd.data.folds import Folds
from chromatinhd.data.fragments import Fragments
from chromatinhd.data.transcriptome import Transcriptome
from chromatinhd.models.pred.model.additive import Models


def zscore(x, dim=0):
    return (x - x.mean(axis=dim, keepdims=True)) / x.std(axis=dim, keepdims=True)


def zscore_relative(x, y, dim=0):
    return (x - y.mean(axis=dim, keepdims=True)) / y.std(axis=dim, keepdims=True)


def fdr(p_vals):
    from scipy.stats import rankdata

    ranked_p_values = rankdata(p_vals)
    qval = p_vals * len(p_vals) / ranked_p_values
    qval[qval > 1] = 1

    return qval


class GenePairWindow(chd.flow.Flow):
    """
    Interpret a *pred* model positionally by censoring windows and comparing
    the decrease in predictivity per cell between pairs of windows
    """

    design = chd.flow.Stored()

    genes = chd.flow.Stored(default=set)

    def score(
        self,
        fragments: Fragments,
        transcriptome: Transcriptome,
        models: Models,
        folds: Folds,
        censorer,
        genes: Optional[List] = None,
        force=False,
        device=None,
    ):
        """
        Score the models

        Parameters:
            fragments:
                the fragments
            transcriptome:
                the transcriptome
            models:
                the models
            folds:
                the folds
            genes:
                which genes to score, defaults to all

        """
        force_ = force
        design = censorer.design.iloc[1:].copy()
        self.design = design

        if device is None:
            device = get_default_device()

        if genes is None:
            genes = transcriptome.var.index

        pbar = tqdm.tqdm(genes, leave=False)

        for gene in pbar:
            pbar.set_description(gene)
            scores_file = self.get_scoring_path(gene) / "scores.pkl"
            interaction_file = self.get_scoring_path(gene) / "interaction.pkl"

            force = force_
            if not all([file.exists() for file in [scores_file, interaction_file]]):
                force = True

            if force:
                deltacor_folds = []
                copredictivity_folds = []
                lost_folds = []
                for fold, model in zip(folds, models):
                    predicted, expected, n_fragments = model.get_prediction_censored(
                        fragments,
                        transcriptome,
                        censorer,
                        cell_ixs=np.concatenate([fold["cells_validation"], fold["cells_test"]]),
                        genes=[gene],
                        device=device,
                    )

                    # select 1st gene, given that we're working with one gene anyway
                    predicted = predicted[..., 0]
                    expected = expected[..., 0]
                    n_fragments = n_fragments[..., 0]

                    # calculate delta cor per cell
                    # calculate effect per cellxgene combination
                    predicted_censored = predicted[1:]
                    predicted_full = predicted[0][None, ...]
                    predicted_full_norm = zscore(predicted_full, 1)
                    predicted_censored_norm = zscore_relative(predicted_censored, predicted_full, 1)

                    expected_norm = zscore(expected[None, ...], 1)

                    celldeltacor = -np.abs(predicted_censored_norm - expected_norm) - -np.abs(
                        predicted_full_norm - expected_norm
                    )
                    with np.errstate(divide="ignore", invalid="ignore"):
                        copredictivity = np.corrcoef(celldeltacor)
                    copredictivity[np.isnan(copredictivity)] = 0.0

                    copredictivity_folds.append(copredictivity)

                    cor = chd.utils.paircor(predicted, expected, dim=-1)
                    deltacor = cor[1:] - cor[0]

                    lost = (n_fragments[0] - n_fragments[1:]).mean(-1)

                    deltacor_folds.append(deltacor)
                    lost_folds.append(lost)

                lost_folds = np.stack(lost_folds, 0)
                deltacor_folds = np.stack(deltacor_folds, 0)
                copredictivity_folds = np.stack(copredictivity_folds, 0)

                result = xr.Dataset(
                    {
                        "deltacor": xr.DataArray(
                            deltacor_folds,
                            coords=[
                                ("model", np.arange(len(models))),
                                ("window", design.index),
                            ],
                        ),
                        "lost": xr.DataArray(
                            lost_folds,
                            coords=[
                                ("model", np.arange(len(models))),
                                ("window", design.index),
                            ],
                        ),
                    }
                )

                windows_oi = lost_folds.mean(0) > 1e-3

                interaction = xr.DataArray(
                    copredictivity_folds[:, windows_oi][:, :, windows_oi],
                    coords=[
                        ("model", np.arange(len(models))),
                        ("window1", design.index[windows_oi]),
                        ("window2", design.index[windows_oi]),
                    ],
                )

                pickle.dump(result, scores_file.open("wb"))
                pickle.dump(interaction, interaction_file.open("wb"))

                self.genes = self.genes | {gene}

    def get_plotdata(self, gene):
        """
        Get plotdata for a gene
        """
        interaction = pickle.load(
            open(
                self.get_scoring_path(gene) / "interaction.pkl",
                "rb",
            )
        )

        plotdata = interaction.mean("model").to_dataframe("cor").reset_index()
        plotdata["window1"] = plotdata["window1"].astype("category")
        plotdata["window2"] = plotdata["window2"].astype("category")

        plotdata = (
            pd.DataFrame(
                itertools.combinations(self.design.index, 2),
                columns=["window1", "window2"],
            )
            .set_index(["window1", "window2"])
            .join(plotdata.set_index(["window1", "window2"]))
        )
        plotdata = plotdata.reset_index().fillna({"cor": 0.0})
        plotdata["window_mid1"] = self.design.loc[plotdata["window1"]]["window_mid"].values
        plotdata["window_mid2"] = self.design.loc[plotdata["window2"]]["window_mid"].values
        plotdata["dist"] = np.abs(plotdata["window_mid1"] - plotdata["window_mid2"])
        plotdata = plotdata.query("(window_mid1 < window_mid2)")
        # plotdata = plotdata.query("dist > 1000")

        # x = interaction.stack({"window1_window2": ["window1", "window2"]}).values
        # print(x.shape)
        # scores_statistical = []
        # for i in range(x.shape[1]):
        #     scores_statistical.append(scipy.stats.ttest_1samp(x[:, i], 0).pvalue)
        # scores_statistical = pd.DataFrame({"pval": scores_statistical})
        # scores_statistical["pval"] = scores_statistical["pval"].fillna(1.0)
        # scores_statistical["qval"] = fdr(scores_statistical["pval"])

        # plotdata["pval"] = scores_statistical["pval"].values
        # plotdata["qval"] = scores_statistical["qval"].values

        plotdata.loc[plotdata["dist"] < 1000, "cor"] = 0.0

        return plotdata

    def get_scoring_path(self, gene):
        path = self.path / f"{gene}"
        path.mkdir(parents=True, exist_ok=True)
        return path



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.7
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Preprocess

# %%
from IPython import get_ipython

if get_ipython():
    get_ipython().run_line_magic("load_ext", "autoreload")
    get_ipython().run_line_magic("autoreload", "2")

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import matplotlib as mpl

import seaborn as sns

sns.set_style("ticks")

import torch

import pickle

import scanpy as sc

import tqdm.auto as tqdm
import io

# %%
import chromatinhd as chd

# %%
folder_root = chd.get_output()
folder_data = folder_root / "data"

dataset_name = "pbmc10k"
main_url = "https://cf.10xgenomics.com/samples/cell-arc/2.0.0/pbmc_granulocyte_sorted_10k/pbmc_granulocyte_sorted_10k"
genome = "GRCh38"

folder_data_preproc = folder_data / dataset_name
folder_data_preproc.mkdir(exist_ok=True, parents=True)

# %% [markdown]
# ## Tiny dataset

# %%
dataset_folder = chd.get_output() / "datasets" / "pbmc10ktiny"
folder_dataset_publish = chd.get_git_root() / "src" / "chromatinhd" / "data" / "examples" / "pbmc10ktiny"
folder_dataset_publish.mkdir(exist_ok=True, parents=True)

# %%
adata = pickle.load((folder_data_preproc / "adata_annotated.pkl").open("rb"))
genes_oi = adata.var.sort_values("dispersions_norm", ascending=False).index[:50]
adata = adata[:, genes_oi]
adata_tiny = sc.AnnData(X=adata[:, genes_oi].layers["magic"], obs=adata.obs, var=adata.var.loc[genes_oi])

# %%
transcriptome = chd.data.transcriptome.Transcriptome.from_adata(adata, path=dataset_folder / "transcriptome")
adata_tiny.write(folder_dataset_publish / "transcriptome.h5ad", compression="gzip")

# %%
selected_transcripts = pickle.load((folder_data_preproc / "selected_transcripts.pkl").open("rb"))
selected_transcripts = selected_transcripts.loc[genes_oi]
regions = chd.data.regions.Regions.from_transcripts(
    selected_transcripts, [-10000, 10000], path=dataset_folder / "regions" / "10k10k"
)

# %%
import pysam

fragments_tabix = pysam.TabixFile(str(folder_data_preproc / "fragments.tsv.gz"))
coordinates = regions.coordinates

fragments_new = []
for i, (gene, promoter_info) in tqdm.tqdm(enumerate(coordinates.iterrows()), total=coordinates.shape[0]):
    start = max(0, promoter_info["start"])

    fragments_promoter = fragments_tabix.fetch(promoter_info["chrom"], start, promoter_info["end"])
    fragments_new.extend(list(fragments_promoter))

fragments_new = pd.DataFrame(
    [x.split("\t") for x in fragments_new], columns=["chrom", "start", "end", "cell", "nreads"]
)
fragments_new["start"] = fragments_new["start"].astype(int)
fragments_new = fragments_new.sort_values(["chrom", "start", "cell"])

# %%
fragments_new.to_csv(folder_dataset_publish / "fragments.tsv", sep="\t", index=False, header=False)
pysam.tabix_compress(folder_dataset_publish / "fragments.tsv", folder_dataset_publish / "fragments.tsv.gz", force=True)

# %%
# !ls -lh {folder_dataset_publish}
# !tabix -p bed {folder_dataset_publish / "fragments.tsv.gz"}
folder_dataset_publish



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.7
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %%
# %load_ext autoreload
# %autoreload 2

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import matplotlib as mpl

import seaborn as sns

sns.set_style("ticks")
# %config InlineBackend.figure_format='retina'

import tqdm.auto as tqdm
import torch
import os
import time

# %%
import chromatinhd as chd

chd.set_default_device("cuda:1")

# %%
dataset_folder_original = chd.get_output() / "datasets" / "pbmc10k"
transcriptome_original = chd.data.Transcriptome(dataset_folder_original / "transcriptome")
fragments_original = chd.data.Fragments(dataset_folder_original / "fragments" / "10k10k")

# %%
genes_oi = transcriptome_original.var.sort_values("dispersions_norm", ascending=False).head(30).index
regions = fragments_original.regions.filter_genes(genes_oi)
fragments = fragments_original.filter_genes(regions)
fragments.create_cellxgene_indptr()
transcriptome = transcriptome_original.filter_genes(regions.coordinates.index)

# %%
folds = chd.data.folds.Folds()
folds.sample_cells(fragments, 5)

# %%
clustering = chd.data.Clustering.from_labels(transcriptome_original.obs["celltype"])

# %%
fold = folds[0]

# %%
models = {}
scores = []

# %%
import logging

logger = chromatinhd.models.diff.trainer.trainer.logger
logger.setLevel(logging.DEBUG)
logger.handlers = []
# logger.handlers = [logging.StreamHandler()]

# %%
devices = pd.DataFrame({"device": ["cuda:1", "cpu"]}).set_index("device")
for device in devices.index:
    if device != "cpu":
        devices.loc[device, "label"] = torch.cuda.get_device_properties(device).name
    else:
        devices.loc[device, "label"] = os.popen("lscpu").read().split("\n")[13].split(": ")[-1].lstrip()

# %%
scores = pd.DataFrame({"device": devices.index}).set_index("device")

# %%
for device in devices.index:
    start = time.time()
    model = chromatinhd.models.diff.model.cutnf.Model(
        fragments,
        clustering,
    )
    model.train_model(fragments, clustering, fold, n_epochs=10, device=device)
    models[device] = model
    end = time.time()
    scores.loc[device, "train"] = end - start

# %%
for device in devices.index:
    genepositional = chromatinhd.models.diff.interpret.genepositional.GenePositional(
        path=chd.get_output() / "interpret" / "genepositional"
    )

    start = time.time()
    genepositional.score(fragments, clustering, [models[device]], force=True, device=device)
    end = time.time()
    scores.loc[device, "inference"] = end - start

# %%
fig = chd.grid.Figure(chd.grid.Wrap(padding_width=0.1))
height = len(scores) * 0.2

plotdata = scores.copy().loc[devices.index]

panel, ax = fig.main.add(chd.grid.Ax((1, height)))
ax.barh(plotdata.index, plotdata["train"])
ax.set_yticks(np.arange(len(devices)))
ax.set_yticklabels(devices.label)
ax.axvline(0, color="black", linestyle="--", lw=1)
ax.set_title("Training")
ax.set_xlabel("seconds")

panel, ax = fig.main.add(chd.grid.Ax((1, height)))
ax.barh(plotdata.index, plotdata["inference"])
ax.axvline(0, color="black", linestyle="--", lw=1)
ax.set_title("Inference")
ax.set_yticks([])
fig.plot()



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.7
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Preprocess

# %%
from IPython import get_ipython

if get_ipython():
    get_ipython().run_line_magic("load_ext", "autoreload")
    get_ipython().run_line_magic("autoreload", "2")

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import matplotlib as mpl

import seaborn as sns

sns.set_style("ticks")

import torch

import pickle

import scanpy as sc

import tqdm.auto as tqdm
import io

# %%
import chromatinhd as chd

# %%
folder_root = chd.get_output()
folder_data = folder_root / "data"

dataset_name = "pbmc10k"
genome = "GRCh38"

folder_data_preproc = folder_data / dataset_name
folder_data_preproc.mkdir(exist_ok=True, parents=True)

# %%
dataset_folder = chd.get_output() / "datasets" / "pbmc10k"
dataset_folder.mkdir(exist_ok=True, parents=True)

# %%
adata = pickle.load((folder_data_preproc / "adata_annotated.pkl").open("rb"))

# %%
transcriptome = chd.data.transcriptome.Transcriptome.from_adata(adata, path=dataset_folder / "transcriptome")

# %%
selected_transcripts = pickle.load((folder_data_preproc / "selected_transcripts.pkl").open("rb"))
regions = chd.data.regions.Regions.from_transcripts(
    selected_transcripts, [-10000, 10000], dataset_folder / "regions" / "10k10k"
)

# %%
fragments_file = folder_data_preproc / "fragments.tsv.gz"
fragments = chd.data.fragments.Fragments(dataset_folder / "fragments" / "10k10k")
fragments.regions = regions
fragments = chd.data.fragments.Fragments.from_fragments_tsv(
    fragments_file=fragments_file,
    regions=regions,
    obs=transcriptome.obs,
    path=fragments.path,
)

# %%



import chromatinhd as chd
import time


class TestAdditive:
    def test_example(self, example_fragments, example_transcriptome, example_folds):
        fold = example_folds[0]
        model = chd.models.pred.model.additive.Model(
            n_genes=len(example_fragments.var.index)
        )

        start = time.time()
        model.train_model(
            example_fragments, example_transcriptome, fold=fold, n_epochs=1
        )

        delta = time.time() - start

        assert delta < 20
        result = model.get_prediction(
            example_fragments, example_transcriptome, cell_ixs=fold["cells_test"]
        )

        assert result["predicted"].shape[0] == len(fold["cells_test"])



import numpy as np
import pandas as pd

import collections
import subprocess as sp
import tqdm.auto as tqdm
import pickle

from chromatinhd.flow import Flow


class PeakCounts(Flow):
    def create_adata(self, original_adata):
        import scanpy as sc

        adata = sc.AnnData(self.counts, obs=self.obs, var=self.var)
        adata.obsm["X_umap"] = original_adata.obsm["X_umap"]
        self.adata = adata

    def count_peaks(self, fragments_location, cell_ids, tabix_location="tabix"):
        # extract unique peaks
        peaks = self.peaks
        unique_peak_ids = list(set(peaks["peak"]))
        peak_idxs = {peak_id: i for i, peak_id in enumerate(unique_peak_ids)}

        # create peaks file for tabix
        self.peaks_bed = peaks[["chrom", "start", "end"]]

        # count
        counts = collections.defaultdict(int)
        cell_ids = [str(cell_id) for cell_id in cell_ids]
        barcode_idxs = {barcode: ix for ix, barcode in enumerate(cell_ids)}

        process = sp.Popen(
            [
                tabix_location,
                fragments_location,
                "-R",
                self.peaks_bed_path,
                "--separate-regions",
            ],
            stdout=sp.PIPE,
        )
        counter = tqdm.tqdm(total=len(unique_peak_ids), smoothing=0)
        for line in process.stdout:
            line = line.decode("utf-8")
            if line.startswith("#"):
                peak = line.rstrip("\n").lstrip("#")
                peak_idx = peak_idxs[peak]
                counter.update(1)
            else:
                fragment = line.split("\t")
                barcode = fragment[3].strip("\n")

                if barcode in barcode_idxs:
                    counts[(barcode_idxs[barcode], peak_idx)] += 1

        # convert to sparse
        import scipy.sparse

        i = [k[0] for k in counts.keys()]
        j = [k[1] for k in counts.keys()]
        v = [v for v in counts.values()]
        counts_csr = scipy.sparse.csr_matrix((v, (i, j)), shape=(len(barcode_idxs), len(peak_idxs)))

        if counts_csr.sum() == 0:
            raise ValueError("Something went wrong with counting")

        self.counts = counts_csr

        # create obs
        obs = pd.DataFrame({"cell": list(barcode_idxs.keys()), "ix": list(barcode_idxs.values())}).set_index("cell")
        self.obs = obs.copy()

        # create var
        var = pd.DataFrame(index=pd.Series(unique_peak_ids, name="peak"))
        var["ix"] = np.arange(var.shape[0])
        self.var = var

    @property
    def peaks_bed_path(self):
        return self.path / "peaks_bed.tsv"

    @property
    def peaks_bed(self):
        return pd.read_table(self.peaks_bed_path)

    @peaks_bed.setter
    def peaks_bed(self, value):
        value.to_csv(self.peaks_bed_path, sep="\t", header=False, index=False)

    @property
    def peaks(self):
        return pd.read_table(self.path / "peaks.tsv", index_col=0)

    @peaks.setter
    def peaks(self, value):
        value.index.name = "peak_gene"
        value.to_csv(self.path / "peaks.tsv", sep="\t")

    @property
    def counts(self):
        return pickle.load((self.path / "counts.pkl").open("rb"))

    @counts.setter
    def counts(self, value):
        pickle.dump(value, (self.path / "counts.pkl").open("wb"))

    @property
    def var(self):
        return pd.read_table(self.path / "var.tsv", index_col=0)

    @var.setter
    def var(self, value):
        value.index.name = "peak"
        value.to_csv(self.path / "var.tsv", sep="\t")

    @property
    def obs(self):
        return pd.read_table(self.path / "obs.tsv", index_col=0)

    @obs.setter
    def obs(self, value):
        value.index.name = "peak"
        value.to_csv(self.path / "obs.tsv", sep="\t")

    _adata = None

    @property
    def adata(self):
        if self._adata is None:
            self._adata = pickle.load((self.path / "adata.pkl").open("rb"))
        return self._adata

    @adata.setter
    def adata(self, value):
        pickle.dump(value, (self.path / "adata.pkl").open("wb"))
        self._adata = value


class FullPeak(PeakCounts):
    default_name = "full_peak"

    def create_peaks(self, original_peak_annot):
        original_peak_annot.index = pd.Index(
            original_peak_annot.chrom
            + ":"
            + original_peak_annot.start.astype(str)
            + "-"
            + original_peak_annot.end.astype(str)
        )

        peaks = original_peak_annot.copy()

        peaks.index = pd.Index(
            peaks.chrom + ":" + peaks.start.astype(str) + "-" + peaks.end.astype(str),
            name="peak",
        )
        peaks = peaks.groupby(level=0).first()
        peaks.index = pd.Index(
            peaks.chrom + ":" + peaks.start.astype(str) + "-" + peaks.end.astype(str),
            name="peak",
        )

        peaks["original_peak"] = peaks.index

        self.peaks = peaks


class HalfPeak(PeakCounts):
    default_name = "half_peak"

    def create_peaks(self, original_peak_annot):
        original_peak_annot.index = pd.Index(
            original_peak_annot.chrom
            + ":"
            + original_peak_annot.start.astype(str)
            + "-"
            + original_peak_annot.end.astype(str)
        )

        peaks = []
        for peak_id, peak in tqdm.tqdm(original_peak_annot.iterrows(), total=original_peak_annot.shape[0]):
            peak1 = peak.copy()
            peak1["end"] = int(peak["start"] + int(peak["end"] - peak["start"]) / 2)
            peak1["original_peak"] = peak_id
            peaks.append(peak1)
            peak2 = peak.copy()
            peak2["start"] = int(peak["end"] - int(peak["end"] - peak["start"]) / 2)
            peak2["original_peak"] = peak_id
            peaks.append(peak2)

        peaks = pd.DataFrame(peaks)
        peaks.index = pd.Index(peaks.chrom + ":" + peaks.start.astype(str) + "-" + peaks.end.astype(str))
        peaks = peaks.groupby(level=0).first()
        peaks.index = pd.Index(peaks.chrom + ":" + peaks.start.astype(str) + "-" + peaks.end.astype(str))

        self.peaks = peaks


class ThirdPeak(PeakCounts):
    default_name = "third_peak"

    def create_peaks(self, original_peak_annot):
        original_peak_annot.index = pd.Index(
            original_peak_annot.chrom
            + ":"
            + original_peak_annot.start.astype(str)
            + "-"
            + original_peak_annot.end.astype(str)
        )

        peaks = []
        for peak_id, peak in original_peak_annot.iterrows():
            scale = peak["end"] - peak["start"]
            peak1 = peak.copy()
            peak1["end"] = int(peak["start"] + scale / 3)
            peak1["original_peak"] = peak_id
            peaks.append(peak1)
            peak2 = peak.copy()
            peak2["start"] = int(peak["start"] + scale / 3)
            peak2["end"] = int(peak["end"] - scale / 3)
            peak2["original_peak"] = peak_id
            peaks.append(peak2)
            peak2 = peak.copy()
            peak2["start"] = int(peak["end"] - scale / 3)
            peak2["original_peak"] = peak_id
            peaks.append(peak2)

        peaks = pd.DataFrame(peaks)
        peaks.index = pd.Index(peaks.chrom + ":" + peaks.start.astype(str) + "-" + peaks.end.astype(str))
        peaks = peaks.groupby(level=0).first()
        peaks.index = pd.Index(peaks.chrom + ":" + peaks.start.astype(str) + "-" + peaks.end.astype(str))

        self.peaks = peaks


class BroaderPeak(PeakCounts):
    default_name = "broader_peak"

    def create_peaks(self, original_peak_annot):
        original_peak_annot.index = pd.Index(
            original_peak_annot.chrom
            + ":"
            + original_peak_annot.start.astype(str)
            + "-"
            + original_peak_annot.end.astype(str)
        )

        peaks = original_peak_annot.copy()
        peaks["original_peak"] = peaks.index.copy()

        peaks["start"] = np.maximum(1, peaks["start"] - 5000)
        peaks["end"] = peaks["end"] + 5000

        peaks.index = pd.Index(peaks.chrom + ":" + peaks.start.astype(str) + "-" + peaks.end.astype(str))
        peaks = peaks.groupby(level=0).first()
        peaks.index = pd.Index(peaks.chrom + ":" + peaks.start.astype(str) + "-" + peaks.end.astype(str))

        self.peaks = peaks



from __future__ import annotations
import math

import numpy as np
import pandas as pd
import torch
import tqdm.auto as tqdm

from chromatinhd.data.regions import Regions
from chromatinhd.flow import TSV, Flow, Linked, Stored, StoredTensor, PathLike
from chromatinhd.utils import class_or_instancemethod

import pathlib


class RawFragments:
    def __init__(self, file):
        self.file = file


class Fragments(Flow):
    """Fragments centered around a gene window"""

    regions: Regions = Linked()
    """The regions where we stored fragments"""

    coordinates: torch.Tensor = StoredTensor(torch.int64)
    """Coordinates of the two cut sites."""

    mapping: torch.Tensor = StoredTensor(torch.int64)
    """Mapping of a fragment to a gene and a cell"""

    cellxgene_indptr: torch.Tensor = StoredTensor(torch.int64)
    """Index pointers for each cellxgene combination"""

    def create_cellxgene_indptr(self):
        cellxgene = self.mapping[:, 0] * self.n_genes + self.mapping[:, 1]

        if not (cellxgene.diff() >= 0).all():
            raise ValueError("Fragments should be ordered by cell then gene (ascending)")

        n_cellxgene = self.n_genes * self.n_cells
        cellxgene_indptr = torch.nn.functional.pad(
            torch.cumsum(torch.bincount(cellxgene, minlength=n_cellxgene), 0), (1, 0)
        )
        assert self.coordinates.shape[0] == cellxgene_indptr[-1]
        if not (cellxgene_indptr.diff() >= 0).all():
            raise ValueError("Fragments should be ordered by cell then gene (ascending)")
        self.cellxgene_indptr = cellxgene_indptr

    _genemapping = None

    @property
    def genemapping(self):
        if self._genemapping is None:
            self._genemapping = self.mapping[:, 1].contiguous()
        return self._genemapping

    _cellmapping = None

    @property
    def cellmapping(self):
        if self._cellmapping is None:
            self._cellmapping = self.mapping[:, 0].contiguous()
        return self._cellmapping

    var = TSV()
    """DataFrame containing information about regions."""

    obs = TSV()
    """DataFrame containing information about cells."""

    _n_genes = None

    @property
    def n_genes(self):
        if self._n_genes is None:
            self._n_genes = self.var.shape[0]
        return self._n_genes

    _n_cells = None

    @property
    def n_cells(self):
        if self._n_cells is None:
            self._n_cells = self.obs.shape[0]
        return self._n_cells

    @property
    def local_cellxgene_ix(self):
        return self.cellmapping * self.n_genes + self.genemapping

    def estimate_fragment_per_cellxgene(self):
        return math.ceil(self.coordinates.shape[0] / self.n_cells / self.n_genes * 2)

    # def create_cut_data(self):
    #     cut_coordinates = self.coordinates.flatten()
    #     cut_coordinates = (cut_coordinates - self.window[0]) / (
    #         self.window[1] - self.window[0]
    #     )
    #     keep_cuts = (cut_coordinates >= 0) & (cut_coordinates <= 1)
    #     cut_coordinates = cut_coordinates[keep_cuts]

    #     self.cut_coordinates = cut_coordinates

    #     self.cut_local_gene_ix = self.genemapping.expand(2, -1).T.flatten()[keep_cuts]
    #     self.cut_local_cell_ix = self.cellmapping.expand(2, -1).T.flatten()[keep_cuts]

    @property
    def genes_oi_torch(self):
        return torch.from_numpy(self.genes_oi).to(self.coordinates.device)

    @property
    def cells_oi_torch(self):
        return torch.from_numpy(self.genes_oi).to(self.coordinates.device)

    @class_or_instancemethod
    def from_fragments_tsv(
        cls,
        fragments_file: PathLike,
        regions: Regions,
        obs: pd.DataFrame,
        cell_column: str = None,
        path: PathLike = None,
        overwrite: bool = True,
    ) -> Fragments:
        """
        Create a Fragments object from a fragments tsv file

        Parameters:
            fragments_file:
                Location of the fragments tab-separate file created by e.g. CellRanger or sinto
            obs:
                DataFrame containing information about cells.
                The index should be the cell names as present in the fragments file.
                Alternatively, the column containing cell ids can be specified using the `cell_column` argument.
            regions:
                Regions from which the fragments will be extracted.
            cell_column:
                Column name in the `obs` DataFrame containing the cell names.
                If not specified, the index of the `obs` DataFrame is used.
            path:
                Folder in which the fragments data will be stored.
            overwrite:
                Whether to overwrite the data if it already exists.
        Returns:
            A new Fragments object
        """

        if isinstance(fragments_file, str):
            fragments_file = pathlib.Path(fragments_file)
        if isinstance(path, str):
            path = pathlib.Path(path)
        if not fragments_file.exists():
            raise FileNotFoundError(f"File {fragments_file} does not exist")
        if not overwrite and path.exists():
            raise FileExistsError(f"Folder {path} already exists")
        path.mkdir(parents=True, exist_ok=True)

        # regions information
        var = pd.DataFrame(index=regions.coordinates.index)
        var["ix"] = np.arange(var.shape[0])

        # cell information
        obs = obs.copy()
        obs["ix"] = np.arange(obs.shape[0])
        if cell_column is None:
            cell_to_cell_ix = obs["ix"].to_dict()
        else:
            cell_to_cell_ix = obs.set_index(cell_column)["ix"].to_dict()

        # load fragments tabix
        import pysam

        fragments_tabix = pysam.TabixFile(str(fragments_file))

        coordinates_raw = []
        mapping_raw = []

        for _, (gene, promoter_info) in tqdm.tqdm(
            enumerate(regions.coordinates.iterrows()),
            total=regions.coordinates.shape[0],
            leave=False,
            desc="Processing fragments",
        ):
            gene_ix = var.loc[gene, "ix"]
            start = max(0, promoter_info["start"])

            fragments_promoter = fragments_tabix.fetch(
                promoter_info["chrom"],
                start,
                promoter_info["end"],
                parser=pysam.asTuple(),
            )

            tss = promoter_info["tss"]
            strand = promoter_info["strand"]

            for fragment in fragments_promoter:
                cell = fragment[3]

                # only store the fragment if the cell is actually of interest
                if cell in cell_to_cell_ix:
                    # add raw data of fragment relative to tss
                    coordinates_raw.append(
                        [
                            (int(fragment[1]) - tss) * strand,
                            (int(fragment[2]) - tss) * strand,
                        ][::strand]
                    )

                    # add mapping of cell/gene
                    mapping_raw.append([cell_to_cell_ix[fragment[3]], gene_ix])

        coordinates = torch.tensor(np.array(coordinates_raw, dtype=np.int64))
        mapping = torch.tensor(np.array(mapping_raw), dtype=torch.int64)

        # Sort `coordinates` and `mapping` according to `mapping`
        sorted_idx = torch.argsort((mapping[:, 0] * var.shape[0] + mapping[:, 1]))
        mapping = mapping[sorted_idx]
        coordinates = coordinates[sorted_idx]

        return cls.create(
            path=path,
            coordinates=coordinates,
            mapping=mapping,
            regions=regions,
            var=var,
            obs=obs,
        )

    def filter_genes(self, regions: Regions, path: PathLike = None) -> Fragments:
        """
        Filter based on new regions

        Parameters:
            regions:
                Regions to filter.
        Returns:
            A new Fragments object
        """

        # test if new regions are a subset of the existing ones
        if not regions.coordinates.index.isin(self.regions.coordinates.index).all():
            raise ValueError("New regions should be a subset of the existing ones")

        # filter genes
        self.regions.coordinates["ix"] = np.arange(self.regions.coordinates.shape[0])
        regions.coordinates["ix"] = self.regions.coordinates["ix"].loc[regions.coordinates.index]
        fragments_oi = np.isin(self.mapping[:, 1].numpy(), regions.coordinates["ix"])

        mapping = self.mapping[fragments_oi]
        coordinates = self.coordinates[fragments_oi]
        var = self.regions.coordinates.copy()
        var["original_ix"] = np.arange(var.shape[0])
        var = var.loc[regions.coordinates.index].copy()
        var["ix"] = np.arange(var.shape[0])
        mapping[:, 1] = torch.from_numpy(var.set_index("original_ix").loc[mapping[:, 1].cpu().numpy(), "ix"].values)

        # Sort `coordinates` and `mapping` according to `mapping`
        sorted_idx = torch.argsort((mapping[:, 0] * var.shape[0] + mapping[:, 1]))
        mapping = mapping[sorted_idx]
        coordinates = coordinates[sorted_idx]

        return Fragments.create(
            coordinates=coordinates, mapping=mapping, regions=regions, var=var, obs=self.obs, path=path
        )


class ChunkedFragments(Flow):
    chunk_size = Stored()
    chunkcoords = StoredTensor(dtype=torch.int64)
    chunkcoords_indptr = StoredTensor(dtype=torch.int32)
    clusters = StoredTensor(dtype=torch.int32)
    relcoords = StoredTensor(dtype=torch.int32)

    clusters = Stored()
    clusters_info = Stored()
    chromosomes = Stored()



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.7
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %%
# %load_ext autoreload
# %autoreload 2

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import matplotlib as mpl

import seaborn as sns

sns.set_style("ticks")
# %config InlineBackend.figure_format='retina'

import tqdm.auto as tqdm

# %%
import chromatinhd as chd

chd.set_default_device("cuda:1")
chd.get_default_device()

# %%
dataset_folder_original = chd.get_output() / "datasets" / "pbmc10k"
transcriptome_original = chd.data.Transcriptome(dataset_folder_original / "transcriptome")
fragments_original = chd.data.Fragments(dataset_folder_original / "fragments" / "10k10k")

# %%
genes_oi = transcriptome_original.var.sort_values("dispersions_norm", ascending=False).head(50).index
regions = fragments_original.regions.filter_genes(genes_oi)
fragments = fragments_original.filter_genes(regions)
fragments.create_cellxgene_indptr()
transcriptome = transcriptome_original.filter_genes(regions.coordinates.index)

# %%
folds = chd.data.folds.Folds()
folds.sample_cells(fragments, 5)

# %%
clustering = chd.data.Clustering.from_labels(transcriptome.obs["celltype"])

# %%
fold = folds[0]

# %% [markdown]
# ## Train

# %%
models = {}

# %%
import logging

logger = chd.models.diff.trainer.trainer.logger
logger.setLevel(logging.DEBUG)
logger.handlers = []
# logger.handlers = [logging.StreamHandler()]

# %%
model = chd.models.diff.model.cutnf.Model(
    fragments,
    clustering,
)
model.train_model(fragments, clustering, fold, n_epochs=30)
models["original"] = model

# %%
model = chd.models.diff.model.cutnf.Model(
    fragments,
    clustering,
)
model.train_model(fragments, clustering, fold, n_epochs=100)
models["original_100epoch"] = model

# %%
model = chd.models.diff.model.cutnf.Model(
    fragments,
    clustering,
    mixture_delta_p_scale=5.0,
)
model.train_model(fragments, clustering, fold, n_epochs=30)
models["original_mixture-delta-p=5"] = model

# %%
model = chd.models.diff.model.cutnf.Model(
    fragments,
    clustering,
    mixture_delta_regularization=False,
    rho_delta_regularization=False,
)
model.train_model(fragments, clustering, fold, n_epochs=30)
models["original_noreg"] = model

# %%
model = chd.models.diff.model.cutnf.Model(
    fragments,
    clustering,
    mixture_delta_p_scale=0.001,
)
model.train_model(fragments, clustering, fold, n_epochs=30)
models["baseline_orig"] = model

# %%
model = chd.models.diff.model.cutnf.Model(
    fragments, clustering, mixture_delta_regularization=False, rho_delta_regularization=False, nbins=(128,)
)
model.train_model(fragments, clustering, fold, n_epochs=30)
models["original_noreg_128"] = model

# %%
model = chd.models.diff.model.cutnf.Model(
    fragments,
    clustering,
    mixture_delta_regularization=False,
    rho_delta_regularization=False,
    nbins=(256,),
)
model.train_model(fragments, clustering, fold, n_epochs=30)
models["original_noreg_256"] = model

# %%
model = chd.models.diff.model.cutnf.Model(
    fragments,
    clustering,
    mixture_delta_regularization=False,
    rho_delta_regularization=False,
    nbins=(256, 128, 64),
)
model.train_model(fragments, clustering, fold, n_epochs=30)
models["original_noreg_256,128,64"] = model

# %%
model = chd.models.diff.model.cutnf.Model(
    fragments,
    clustering,
    nbins=(256, 128, 64),
)
model.train_model(fragments, clustering, fold, n_epochs=30)
models["original_256,128,64"] = model

# %%
model = chd.models.diff.model.cutnf.Model(
    fragments,
    clustering,
    nbins=(512, 256, 128, 64),
)
model.train_model(fragments, clustering, fold, n_epochs=30)
models["original_512,256,128,64"] = model

# %%
model = chd.models.diff.model.cutnf.Model(fragments, clustering, mixture_delta_p_scale=0.001, nbins=(128,))
model.train_model(fragments, clustering, fold, n_epochs=30)
models["baseline_128"] = model

# %%
model = chd.models.diff.model.cutnf.Model(fragments, clustering, mixture_delta_p_scale=0.001, nbins=(256,))
model.train_model(fragments, clustering, fold, n_epochs=30)
models["baseline_256"] = model

# %%
model = chd.models.diff.model.cutnf.Model(fragments, clustering, mixture_delta_p_scale=0.001, nbins=(256, 128, 64))
model.train_model(fragments, clustering, fold, n_epochs=30)
models["baseline_256,128,64"] = model

# %% [markdown]
# ## Score

# %%
scores = []
genescores = []
for model_id, model in models.items():
    prediction_test = model.get_prediction(fragments, clustering, cell_ixs=fold["cells_test"])
    prediction_validation = model.get_prediction(fragments, clustering, cell_ixs=fold["cells_validation"])
    prediction_train = model.get_prediction(fragments, clustering, cell_ixs=fold["cells_train"])
    scores.append(
        {
            "model_id": model_id,
            "lik_test": (prediction_test["likelihood_mixture"]).sum().item(),
            "n_test": len(fold["cells_test"]),
            "lik_validation": (prediction_validation["likelihood_mixture"]).sum().item(),
            "n_validation": len(fold["cells_validation"]),
            "lik_train": (prediction_train["likelihood_mixture"]).sum().item(),
            "n_train": len(fold["cells_train"]),
        }
    )
    genescores.append(
        pd.DataFrame(
            {
                "model_id": model_id,
                "lik_test": (prediction_test["likelihood_mixture"]).sum("cell").to_pandas(),
                "n_test": len(fold["cells_test"]),
                "lik_validation": (prediction_validation["likelihood_mixture"]).sum("cell").to_pandas(),
                "n_validation": len(fold["cells_validation"]),
                "lik_train": (prediction_train["likelihood_mixture"]).sum("cell").to_pandas(),
                "n_train": len(fold["cells_train"]),
            }
        )
    )
scores = pd.DataFrame(scores).set_index("model_id")
genescores = (
    pd.concat([genescores[i] for i in range(len(genescores))], axis=0).reset_index().set_index(["model_id", "gene"])
)

# %%
baseline_id = "baseline_orig"
scores[["lr_test", "lr_validation", "lr_train"]] = (
    scores[["lik_test", "lik_validation", "lik_train"]]
    - scores[["lik_test", "lik_validation", "lik_train"]].loc[baseline_id]
)
scores[["nlr_test", "nlr_validation", "nlr_train"]] = (
    scores[["lr_test", "lr_validation", "lr_train"]].values / scores[["n_test", "n_validation", "n_train"]].values
)
genescores[["lr_test", "lr_validation", "lr_train"]] = (
    genescores[["lik_test", "lik_validation", "lik_train"]]
    - genescores[["lik_test", "lik_validation", "lik_train"]].loc[baseline_id]
)
genescores[["nlr_test", "nlr_validation", "nlr_train"]] = (
    genescores[["lr_test", "lr_validation", "lr_train"]].values
    / genescores[["n_test", "n_validation", "n_train"]].values
)

# %%
model_info = pd.DataFrame({"model": models.keys()}).set_index("model")
model_info["model_type"] = model_info.index.map(lambda x: "_".join(x.split("_")[:-1]))
model_info = model_info.sort_values(["model_type"])
model_info["ix"] = np.arange(model_info.shape[0])

# %%
fig = chd.grid.Figure(chd.grid.Wrap(padding_width=0.1))
height = len(scores) * 0.2

plotdata = scores.copy().loc[model_info.index]

panel, ax = fig.main.add(chd.grid.Ax((1, height)))
ax.barh(plotdata.index, plotdata["lr_test"])
ax.axvline(0, color="black", linestyle="--", lw=1)
ax.set_title("Test")
ax.set_xlabel("Log-likehood ratio")

panel, ax = fig.main.add(chd.grid.Ax((1, height)))
ax.set_yticks([])
ax.barh(plotdata.index, plotdata["lr_validation"])
ax.axvline(0, color="black", linestyle="--", lw=1)
ax.set_title("Validation")

panel, ax = fig.main.add(chd.grid.Ax((1, height)))
ax.set_yticks([])
ax.barh(plotdata.index, plotdata["lr_train"])
ax.axvline(0, color="black", linestyle="--", lw=1)
ax.set_title("Train")
fig.plot()

# %%
plotdata = genescores["lr_test"].unstack()
plotdata.columns = transcriptome.symbol(plotdata.columns)
plotdata = plotdata.loc[model_info.index].T

fig, ax = plt.subplots(figsize=(plotdata.shape[1] * 0.2, plotdata.shape[0] * 0.2))
sns.heatmap(plotdata, vmax=100, vmin=-100, cmap="RdBu_r", center=0, cbar_kws={"shrink": 0.5}, yticklabels=True)

# add dot for highest
for i, gene in enumerate(plotdata.index):
    j = plotdata.loc[gene].argmax()
    plt.plot(j + 0.5, i + 0.5, "o", color="black", markersize=4, markeredgewidth=1.0, markeredgecolor="white")

# %% [markdown]
# ## Interpret

# %%
genepositional = chd.models.diff.interpret.genepositional.GenePositional(
    path=chd.get_output() / "interpret" / "genepositional"
)

# %%
symbol = "EBF1"
model_id = "original"
# model_id = "original_512,256,128,64"

genepositional.score(fragments, clustering, [models[model_id]], force=True, genes=transcriptome.gene_id([symbol]))

# %%
fig = chd.grid.Figure(chd.grid.Grid(padding_height=0.05, padding_width=0.05))
width = 10

region = fragments.regions.coordinates.loc[transcriptome.gene_id(symbol)]
panel_genes = chd.plot.genome.genes.Genes.from_region(region, width=width)
fig.main.add_under(panel_genes)

plotdata, plotdata_mean = genepositional.get_plotdata(transcriptome.gene_id(symbol))
panel_differential = chd.models.diff.plot.Differential(
    plotdata, plotdata_mean, cluster_info=clustering.cluster_info, panel_height=0.5, width=width
)
fig.main.add_under(panel_differential)

panel_expression = chd.models.diff.plot.DifferentialExpression.from_transcriptome(
    transcriptome=transcriptome, clustering=clustering, gene=transcriptome.gene_id(symbol), panel_height=0.5
)
fig.main.add_right(panel_expression, row=panel_differential)

fig.plot()



def paircor(x, y, dim=0, eps=0.1):
    divisor = (y.std(dim) * x.std(dim)) + eps
    cor = ((x - x.mean(dim, keepdims=True)) * (y - y.mean(dim, keepdims=True))).mean(
        dim
    ) / divisor
    return cor


def paircor_loss(x, y):
    return -paircor(x, y).mean() * 100


def gene_paircor_loss(x, y):
    return -paircor(x, y) * 100



import math
import pickle

import numpy as np
import torch
import torch_scatter
import tqdm.auto as tqdm
import xarray as xr
import pandas as pd

from chromatinhd.embedding import EmbeddingTensor
from chromatinhd.flow import Flow, Stored
from chromatinhd.loaders import LoaderPool
from chromatinhd.models import HybridModel
from chromatinhd.models.diff.loader.clustering_cuts import ClusteringCuts
from chromatinhd.models.diff.loader.minibatches import Minibatcher
from chromatinhd.models.diff.trainer import Trainer
from chromatinhd.optim import SparseDenseAdam
from chromatinhd.data.fragments import Fragments
from chromatinhd.data.clustering import Clustering
from chromatinhd.utils import crossing
from chromatinhd import get_default_device
from .spline import DifferentialQuadraticSplineStack, TransformedDistribution


class Decoder(torch.nn.Module):
    def __init__(
        self,
        n_latent,
        n_genes,
        n_output_components,
        n_layers=1,
        n_hidden_dimensions=32,
        dropout_rate=0.2,
    ):
        super().__init__()

        layers = []
        for i in range(n_layers):
            layers.append(torch.nn.Linear(n_hidden_dimensions if i > 0 else n_latent, n_hidden_dimensions))
            layers.append(torch.nn.BatchNorm1d(n_hidden_dimensions))
            layers.append(torch.nn.ReLU())
            if dropout_rate > 0.0:
                layers.append(torch.nn.Dropout(0.2))
        self.nn = torch.nn.Sequential(*layers)

        self.n_hidden_dimensions = n_hidden_dimensions
        self.n_output_components = n_output_components

        self.logit_weight = EmbeddingTensor(
            n_genes,
            (n_hidden_dimensions if n_layers > 0 else n_latent, n_output_components),
            sparse=True,
        )
        stdv = 1.0 / math.sqrt(self.logit_weight.weight.size(1))
        # self.logit_weight.weight.data.uniform_(-stdv, stdv)
        self.logit_weight.weight.data.zero_()

        self.rho_weight = EmbeddingTensor(n_genes, (n_hidden_dimensions if n_layers > 0 else n_latent,), sparse=True)
        stdv = 1.0 / math.sqrt(self.rho_weight.weight.size(1))
        self.rho_weight.weight.data.uniform_(-stdv, stdv)

    def forward(self, latent, genes_oi):
        # genes oi is only used to get the logits
        # we calculate the rho for all genes because we pool using softmax later
        logit_weight = self.logit_weight(genes_oi)
        rho_weight = self.rho_weight.get_full_weight()
        nn_output = self.nn(latent)

        # nn_output is broadcasted across genes and across components
        logit = torch.matmul(nn_output.unsqueeze(1).unsqueeze(2), logit_weight).squeeze(-2)

        # nn_output has to be broadcasted across genes
        rho = torch.matmul(nn_output.unsqueeze(1), rho_weight.T).squeeze(-2)

        return logit, rho

    def parameters_dense(self):
        return [*self.nn.parameters(), self.rho_weight.weight]

    def parameters_sparse(self):
        return [self.logit_weight.weight]


class BaselineDecoder(torch.nn.Module):
    def __init__(self, n_latent, n_genes, n_output_components, n_layers=1, n_hidden_dimensions=32):
        super().__init__()

        layers = []
        for i in range(n_layers):
            layers.append(torch.nn.Linear(n_hidden_dimensions if i > 0 else n_latent, n_hidden_dimensions))
            layers.append(torch.nn.BatchNorm1d(n_hidden_dimensions))
            layers.append(torch.nn.ReLU())
            layers.append(torch.nn.Dropout(0.2))
        self.nn = torch.nn.Sequential(*layers)

        self.n_hidden_dimensions = n_hidden_dimensions
        self.n_output_components = n_output_components

        self.rho_weight = EmbeddingTensor(n_genes, (n_hidden_dimensions if n_layers > 0 else n_latent,), sparse=True)
        stdv = 1.0 / math.sqrt(self.rho_weight.weight.size(1)) / 100
        self.rho_weight.weight.data.uniform_(-stdv, stdv)

    def forward(self, latent, genes_oi):
        rho_weight = self.rho_weight(genes_oi)
        nn_output = self.nn(latent)

        # nn_output is broadcasted across genes and across components
        logit = torch.zeros(
            (latent.shape[0], len(genes_oi), self.n_output_components),
            device=latent.device,
        )

        rho = torch.matmul(nn_output, rho_weight.T).squeeze(-1)
        return logit, rho

    def parameters_dense(self):
        return [*self.nn.parameters()]

    def parameters_sparse(self):
        return [self.rho_weight.weight]


class Model(torch.nn.Module, HybridModel):
    """
    A ChromatinHD-diff model that models the probability density of observing a cut site between clusterings
    """

    def __init__(
        self,
        fragments,
        clustering,
        nbins=(
            128,
            64,
            32,
        ),
        decoder_n_layers=0,
        baseline=False,
        scale_likelihood=False,
        rho_delta_regularization=True,
        rho_delta_p_scale_free=False,
        mixture_delta_regularization=True,
        mixture_delta_p_scale_free=False,
        mixture_delta_p_scale_dist="normal",
        mixture_delta_p_scale=1.0,
    ):
        super().__init__()

        self.n_total_genes = fragments.n_genes

        self.n_clusters = clustering.n_clusters

        transform = DifferentialQuadraticSplineStack(
            nbins=nbins,
            n_genes=fragments.n_genes,
        )
        self.mixture = TransformedDistribution(transform)
        n_delta_mixture_components = sum(transform.split_deltas)

        if not baseline:
            self.decoder = Decoder(
                self.n_clusters,
                fragments.n_genes,
                n_delta_mixture_components,
                n_layers=decoder_n_layers,
            )
        else:
            self.decoder = BaselineDecoder(
                self.n_clusters,
                fragments.n_genes,
                n_delta_mixture_components,
                n_layers=decoder_n_layers,
            )

        libsize = torch.bincount(fragments.mapping[:, 0], minlength=fragments.n_cells)
        rho_bias = (
            torch.bincount(fragments.mapping[:, 1], minlength=fragments.n_genes)
            / fragments.n_cells
            / libsize.to(torch.float).mean()
        )
        min_rho_bias = 1e-5
        rho_bias = min_rho_bias + (1 - min_rho_bias) * rho_bias

        self.register_buffer("libsize", libsize)
        self.register_buffer("rho_bias", rho_bias)

        self.track = {}

        self.mixture_delta_regularization = mixture_delta_regularization
        if self.mixture_delta_regularization:
            if mixture_delta_p_scale_free:
                self.mixture_delta_p_scale = torch.nn.Parameter(
                    torch.tensor(math.log(mixture_delta_p_scale), requires_grad=True)
                )
            else:
                self.register_buffer(
                    "mixture_delta_p_scale",
                    torch.tensor(math.log(mixture_delta_p_scale)),
                )

        self.n_total_cells = fragments.n_cells

        self.scale_likelihood = scale_likelihood

        self.rho_delta_regularization = rho_delta_regularization
        if self.rho_delta_regularization:
            if rho_delta_p_scale_free:
                self.rho_delta_p_scale = torch.nn.Parameter(torch.log(torch.tensor(0.1, requires_grad=True)))
            else:
                self.register_buffer("rho_delta_p_scale", torch.tensor(math.log(1.0)))

        self.mixture_delta_p_scale_dist = mixture_delta_p_scale_dist

    def forward_(
        self,
        coordinates,
        clustering,
        genes_oi,
        local_cellxgene_ix,
        localcellxgene_ix,
        local_gene_ix,
    ):
        # decode
        mixture_delta, rho_delta = self.decoder(clustering.to(torch.float), genes_oi)

        # rho
        rho = torch.nn.functional.softmax(torch.log(self.rho_bias) + rho_delta, -1)
        rho_cuts = rho.flatten()[localcellxgene_ix]

        # fragment counts
        mixture_delta_cellxgene = mixture_delta.view(np.prod(mixture_delta.shape[:2]), mixture_delta.shape[-1])
        mixture_delta = mixture_delta_cellxgene[local_cellxgene_ix]

        self.track["likelihood_mixture"] = likelihood_mixture = self.mixture.log_prob(
            coordinates, genes_oi, local_gene_ix, mixture_delta
        )

        self.track["likelihood_overall"] = likelihood_overall = torch.log(rho_cuts) + math.log(self.n_total_genes)

        # overall likelihood
        likelihood = self.track["likelihood"] = likelihood_mixture + likelihood_overall
        likelihood_scale = 1.0

        elbo = -likelihood.sum() * likelihood_scale

        # regularization
        # mixture
        if self.mixture_delta_regularization:
            mixture_delta_p = torch.distributions.Normal(0.0, torch.exp(self.mixture_delta_p_scale))
            mixture_delta_kl = mixture_delta_p.log_prob(self.decoder.logit_weight(genes_oi))

            elbo -= mixture_delta_kl.sum()

        # rho delta
        if self.rho_delta_regularization:
            rho_delta_p = torch.distributions.Normal(0.0, torch.exp(self.rho_delta_p_scale))
            rho_delta_kl = rho_delta_p.log_prob(self.decoder.rho_weight(genes_oi))

            elbo -= rho_delta_kl.sum()

        return elbo

    def forward(self, data):
        return self.forward_(
            coordinates=data.cuts.coordinates,
            clustering=data.clustering.onehot,
            genes_oi=data.minibatch.genes_oi_torch,
            local_gene_ix=data.cuts.local_gene_ix,
            local_cellxgene_ix=data.cuts.local_cellxgene_ix,
            localcellxgene_ix=data.cuts.localcellxgene_ix,
        )

    def train_model(self, fragments, clustering, fold, device=None, n_epochs=30, lr=1e-2):
        """
        Trains the model
        """

        if device is None:
            device = get_default_device()

        # set up minibatchers and loaders
        minibatcher_train = Minibatcher(
            fold["cells_train"],
            range(fragments.n_genes),
            n_genes_step=500,
            n_cells_step=200,
        )
        minibatcher_validation = Minibatcher(
            fold["cells_validation"],
            range(fragments.n_genes),
            n_genes_step=10,
            n_cells_step=10000,
            permute_cells=False,
            permute_genes=False,
        )

        loaders_train = LoaderPool(
            ClusteringCuts,
            dict(
                clustering=clustering,
                fragments=fragments,
                cellxgene_batch_size=minibatcher_train.cellxgene_batch_size,
            ),
            n_workers=10,
        )
        loaders_validation = LoaderPool(
            ClusteringCuts,
            dict(
                clustering=clustering,
                fragments=fragments,
                cellxgene_batch_size=minibatcher_validation.cellxgene_batch_size,
            ),
            n_workers=5,
        )

        trainer = Trainer(
            self,
            loaders_train,
            loaders_validation,
            minibatcher_train,
            minibatcher_validation,
            SparseDenseAdam(
                self.parameters_sparse(),
                self.parameters_dense(),
                lr=lr,
                weight_decay=1e-5,
            ),
            n_epochs=n_epochs,
            checkpoint_every_epoch=1,
            optimize_every_step=1,
            device=device,
        )
        self.trace = trainer.trace

        trainer.train()

    def _get_likelihood_cell_gene(self, likelihood, local_cellxgene_ix, n_cells, n_genes):
        return torch_scatter.segment_sum_coo(likelihood, local_cellxgene_ix, dim_size=n_cells * n_genes).reshape(
            (n_cells, n_genes)
        )

    def get_prediction(
        self,
        fragments: Fragments,
        clustering: Clustering,
        cells=None,
        cell_ixs=None,
        genes=None,
        gene_ixs=None,
        device: str = None,
    ):
        """
        Returns the prediction of a dataset

        Parameters:
            fragments: Fragments object
            clustering: Clustering object
            cells: Cells to predict
            cell_ixs: Cell indices to predict
            genes: Genes to predict
            gene_ixs: Gene indices to predict
            device: Device to use
        """
        if cell_ixs is None:
            if cells is None:
                cells = fragments.obs.index
            fragments.obs["ix"] = np.arange(len(fragments.obs))
            cell_ixs = fragments.obs.loc[cells]["ix"].values
        if cells is None:
            cells = fragments.obs.index[cell_ixs]

        if gene_ixs is None:
            if genes is None:
                genes = fragments.var.index
            fragments.var["ix"] = np.arange(len(fragments.var))
            gene_ixs = fragments.var.loc[genes]["ix"].values
        if genes is None:
            genes = fragments.var.index[gene_ixs]

        minibatches = Minibatcher(
            cell_ixs,
            gene_ixs,
            n_genes_step=500,
            n_cells_step=200,
            use_all_cells=True,
            use_all_genes=True,
            permute_cells=False,
            permute_genes=False,
        )
        loaders = LoaderPool(
            ClusteringCuts,
            dict(
                clustering=clustering,
                fragments=fragments,
                cellxgene_batch_size=minibatches.cellxgene_batch_size,
            ),
            n_workers=5,
        )
        loaders.initialize(minibatches)

        likelihood_mixture = np.zeros((len(cell_ixs), len(gene_ixs)))
        likelihood_overall = np.zeros((len(cell_ixs), len(gene_ixs)))

        cell_mapping = np.zeros(fragments.n_cells, dtype=np.int64)
        cell_mapping[cell_ixs] = np.arange(len(cell_ixs))

        gene_mapping = np.zeros(fragments.n_genes, dtype=np.int64)
        gene_mapping[gene_ixs] = np.arange(len(gene_ixs))

        self.eval()
        self = self.to(device)

        for data in loaders:
            data = data.to(device)
            with torch.no_grad():
                self.forward(data)

            likelihood_mixture[
                np.ix_(
                    cell_mapping[data.minibatch.cells_oi],
                    gene_mapping[data.minibatch.genes_oi],
                )
            ] += (
                self._get_likelihood_cell_gene(
                    self.track["likelihood_mixture"],
                    data.cuts.local_cellxgene_ix,
                    data.minibatch.n_cells,
                    data.minibatch.n_genes,
                )
                .cpu()
                .numpy()
            )
            likelihood_overall[
                np.ix_(
                    cell_mapping[data.minibatch.cells_oi],
                    gene_mapping[data.minibatch.genes_oi],
                )
            ] += (
                self._get_likelihood_cell_gene(
                    self.track["likelihood_overall"],
                    data.cuts.local_cellxgene_ix,
                    data.minibatch.n_cells,
                    data.minibatch.n_genes,
                )
                .cpu()
                .numpy()
            )

        self = self.to("cpu")

        result = xr.Dataset(
            {
                "likelihood_mixture": xr.DataArray(
                    likelihood_mixture,
                    dims=("cell", "gene"),
                    coords={"cell": cells, "gene": fragments.var.index},
                ),
                "likelihood_overall": xr.DataArray(
                    likelihood_overall,
                    dims=("cell", "gene"),
                    coords={"cell": cells, "gene": fragments.var.index},
                ),
            }
        )
        return result

    def evaluate_pseudo(
        self,
        coordinates,
        clustering=None,
        gene_oi=None,
        gene_ix=None,
        device=None,
    ):
        from chromatinhd.models.diff.loader.clustering import Result as ClusteringResult
        from chromatinhd.models.diff.loader.clustering_cuts import (
            Result as ClusteringCutsResult,
        )
        from chromatinhd.models.diff.loader.cuts import Result as CutsResult
        from chromatinhd.models.diff.loader.minibatches import Minibatch

        if not torch.is_tensor(clustering):
            if clustering is None:
                clustering = 0.0
            clustering = torch.ones((1, self.n_clusters)) * clustering

            print(clustering)

        cells_oi = torch.ones((1,), dtype=torch.long)

        local_cellxgene_ix = torch.tensor([], dtype=torch.long)
        if gene_ix is None:
            if gene_oi is None:
                gene_oi = 0
            genes_oi = torch.tensor([gene_oi], dtype=torch.long)
            local_gene_ix = torch.zeros_like(coordinates).to(torch.long)
            local_cellxgene_ix = torch.zeros_like(coordinates).to(torch.long)
            localcellxgene_ix = torch.ones_like(coordinates).to(torch.long) * gene_oi
        else:
            assert len(gene_ix) == len(coordinates)
            genes_oi = torch.unique(gene_ix)

            local_gene_mapping = torch.zeros(genes_oi.max() + 1, dtype=torch.long)
            local_gene_mapping.index_add_(0, genes_oi, torch.arange(len(genes_oi)))

            local_gene_ix = local_gene_mapping[gene_ix]
            local_cell_ix = torch.arange(clustering.shape[0])
            local_cellxgene_ix = local_cell_ix * len(genes_oi) + local_gene_ix
            localcellxgene_ix = local_cell_ix * self.n_total_genes + gene_ix

        data = ClusteringCutsResult(
            cuts=CutsResult(
                coordinates=coordinates,
                local_cellxgene_ix=local_cellxgene_ix,
                localcellxgene_ix=localcellxgene_ix,
                n_genes=len(genes_oi),
            ),
            clustering=ClusteringResult(
                onehot=clustering,
            ),
            minibatch=Minibatch(
                cells_oi=cells_oi.cpu().numpy(),
                genes_oi=genes_oi.cpu().numpy(),
            ),
        ).to(device)

        self = self.to(device).eval()

        with torch.no_grad():
            self.forward(data)

        self = self.to("cpu")

        prob = self.track["likelihood"].detach().cpu()
        return prob.detach().cpu()

    def rank(
        self,
        window: np.ndarray,
        n_latent: int,
        device: torch.DeviceObjType = None,
        how: str = "probs_diff_masked",
        prob_cutoff: float = None,
    ) -> np.ndarray:
        n_genes = self.rho_bias.shape[0]

        self = self.to(device).eval()

        # create design for inference
        design_gene = pd.DataFrame({"gene_ix": np.arange(n_genes)})
        design_latent = pd.DataFrame({"active_latent": np.arange(n_latent)})
        design_coord = pd.DataFrame({"coord": np.arange(window[0], window[1] + 1, step=25)})
        design = crossing(design_gene, design_latent, design_coord)
        batch_size = 100000
        design["batch"] = np.floor(np.arange(design.shape[0]) / batch_size).astype(int)

        # infer
        probs = []
        for _, design_subset in tqdm.tqdm(design.groupby("batch")):
            pseudocoordinates = torch.from_numpy(design_subset["coord"].values).to(device)
            pseudocoordinates = (pseudocoordinates - window[0]) / (window[1] - window[0])
            pseudolatent = torch.nn.functional.one_hot(
                torch.from_numpy(design_subset["active_latent"].values).to(device),
                n_latent,
            ).to(torch.float)
            gene_ix = torch.from_numpy(design_subset["gene_ix"].values).to(device)

            prob = self.evaluate_pseudo(
                pseudocoordinates.to(device),
                latent=pseudolatent.to(device),
                gene_ix=gene_ix,
            )

            probs.append(prob)
        probs = np.hstack(probs)

        probs = probs.reshape((design_gene.shape[0], design_latent.shape[0], design_coord.shape[0]))

        # calculate the score we're gonna use: how much does the likelihood of
        # a cut in a window change compared to the "mean"?
        probs_diff = probs - probs.mean(-2, keepdims=True)

        # apply a mask to regions with very low likelihood of a cut
        if prob_cutoff is None:
            prob_cutoff = np.log(1.0)
        mask = probs >= prob_cutoff

        probs_diff_masked = probs_diff.copy()
        probs_diff_masked[~mask] = -np.inf

        ## Single base-pair resolution
        # interpolate the scoring from above but now at single base pairs
        # we may have to smooth this in the future,
        # particularly for very detailed models that already look at base pair resolution
        x = (design["coord"].values).reshape((design_gene.shape[0], design_latent.shape[0], design_coord.shape[0]))

        def interpolate(x: torch.Tensor, xp: torch.Tensor, fp: torch.Tensor) -> torch.Tensor:
            a = (fp[..., 1:] - fp[..., :-1]) / (xp[..., 1:] - xp[..., :-1])
            b = fp[..., :-1] - (a.mul(xp[..., :-1]))

            indices = torch.searchsorted(xp.contiguous(), x.contiguous(), right=False) - 1
            indices = torch.clamp(indices, 0, a.shape[-1] - 1)
            slope = a.index_select(a.ndim - 1, indices)
            intercept = b.index_select(a.ndim - 1, indices)
            return x * slope + intercept

        desired_x = torch.arange(*window)

        probs_diff_interpolated = interpolate(
            desired_x, torch.from_numpy(x)[0][0], torch.from_numpy(probs_diff)
        ).numpy()
        probs_interpolated = interpolate(desired_x, torch.from_numpy(x)[0][0], torch.from_numpy(probs)).numpy()

        if how == "probs_diff_masked":
            # again apply a mask
            probs_diff_interpolated_masked = probs_diff_interpolated.copy()
            mask_interpolated = probs_interpolated >= prob_cutoff
            probs_diff_interpolated_masked[~mask_interpolated] = -np.inf

            return probs_diff_interpolated_masked
        else:
            raise NotImplementedError()


class Models(Flow):
    n_models = Stored()

    @property
    def models_path(self):
        path = self.path / "models"
        path.mkdir(exist_ok=True)
        return path

    def train_models(self, fragments, clustering, folds, device=None, n_epochs=30, **kwargs):
        """
        Trains the models

        Parameters:
            fragments:
                Fragments object
        """
        self.n_models = len(folds)
        for fold_ix, fold in [(fold_ix, fold) for fold_ix, fold in enumerate(folds)]:
            desired_outputs = [self.models_path / ("model_" + str(fold_ix) + ".pkl")]
            force = False
            if not all([desired_output.exists() for desired_output in desired_outputs]):
                force = True

            if force:
                model = Model(fragments, clustering, **kwargs)
                model.train_model(fragments, clustering, fold, device=device, n_epochs=n_epochs)

                model = model.to("cpu")

                pickle.dump(
                    model,
                    open(self.models_path / ("model_" + str(fold_ix) + ".pkl"), "wb"),
                )

    def __getitem__(self, ix):
        return pickle.load((self.models_path / ("model_" + str(ix) + ".pkl")).open("rb"))

    def __len__(self):
        return self.n_models

    def __iter__(self):
        for ix in range(len(self)):
            yield self[ix]



import pathlib
from typing import Union

import numpy as np
import pandas as pd
import torch
import tqdm.auto as tqdm

from chromatinhd import get_default_device
from chromatinhd.data.regions import Regions
from chromatinhd.flow import (
    CompressedNumpyFloat64,
    CompressedNumpyInt64,
    Flow,
    Linked,
    Stored,
    StoredDataFrame,
)
from chromatinhd.utils.torch import ind2ptr


class Motifscan(Flow):
    """
    A sprase representation of locations of different motifs in regions of the genome
    """

    regions = Linked()
    "The regions"

    indptr = CompressedNumpyInt64()
    "The index pointers for each position in the regions"

    position = CompressedNumpyInt64()
    "Position associated to each site"

    indices = CompressedNumpyInt64()
    "Motif index associated to each site"

    scores = CompressedNumpyFloat64()
    "Scores associated with each detected site"

    strands = CompressedNumpyFloat64()
    "Strand associated with each detected site"

    shape = Stored()

    n_motifs = Stored()
    "Number of motifs"

    motifs = StoredDataFrame()
    "Dataframe storing auxilliary information for each motif"

    @classmethod
    def from_pwms(
        cls,
        pwms: dict,
        regions: Regions,
        fasta_file: Union[str, pathlib.Path],
        path: Union[str, pathlib.Path],
        cutoffs: Union[int, float, pd.Series] = None,
        cutoff_col: str = None,
        motifs: pd.DataFrame = None,
        device=None,
        batch_size: int = 5000000,
    ):
        """
        Create a motifscan object from a set of pwms and a set of regions

        Parameters:
            pwms:
                A dictionary of pwms, where the keys are the motif ids and the values are the pwms
            regions:
                A regions object
            fasta_file:
                The location of the fasta file containing the genome
            motifs:
                A dataframe containing auxilliary information for each motif
            path:
                The folder where the motifscan data will be stored.
            cutoffs:
                A dictionary containing the cutoffs for each motif.
            cutoff_col:
                The column in the motifs dataframe containing the cutoffs
            device:
                The device to use for the scanning
            batch_size:
                The batch size to use for scanning. Lower batch size if the GPU runs out of memory
        """

        if device is None:
            device = get_default_device()

        self = cls(path)

        # check or create cutoffs
        if cutoffs is None:
            if cutoff_col is None:
                raise ValueError("Either motifs+cutoff_col or cutoffs need to be specified.")
            if motifs is None:
                raise ValueError("Either motifs+cutoff_col or cutoffs need to be specified. motifs is not given")

            cutoffs = motifs[cutoff_col].to_dict()
        else:
            if isinstance(cutoffs, (float, int)):
                cutoffs = {motif: cutoffs for motif in pwms.keys()}
            elif isinstance(cutoffs, pd.Series):
                cutoffs = cutoffs.to_dict()
            else:
                raise ValueError("cutoffs should be a float, int, dict or pd.Series")
            assert set(cutoffs.keys()) == set(pwms.keys())

        # check or create motifs
        if motifs is None:
            motifs = pd.DataFrame(
                {
                    "motif": list(pwms.keys()),
                }
            ).set_index("motif")

        # divide regions into batches according to batch size
        region_coordinates = regions.coordinates

        region_coordinates = divide_regions_in_batches(region_coordinates, batch_size=batch_size)

        region_size = region_coordinates["end"].values[0] - region_coordinates["start"].values[0]

        # load in fasta file
        import pysam

        fasta = pysam.FastaFile(fasta_file)

        # do the actual counting by looping over the batches, extract the sequences and scan
        positions = []
        indices = []
        scores = []
        strands = []

        for batch, region_coordinates_batch in tqdm.tqdm(region_coordinates.groupby("batch")):
            sequences = [
                fasta.fetch(chrom, start, end + 1)
                for chrom, start, end in region_coordinates_batch[["chrom", "start", "end"]].values
            ]
            assert (
                len(set(len(sequence) for sequence in sequences)) == 1
            ), "All regions/sequences should have the same length"
            onehot = torch.stack([create_onehot(digitize_sequence(sequence)) for sequence in sequences]).to(device)
            for motif_ix, motif in enumerate(motifs.index):
                cutoff = cutoffs[motif]

                # get pwm
                pwm = pwms[motif]
                if not torch.is_tensor(pwm):
                    pwm = torch.from_numpy(pwm)
                pwm = pwm.to(dtype=torch.float32, device=onehot.device)

                (
                    scores_motif,
                    positions_motif,
                    strands_motif,
                ) = scan(onehot, pwm, cutoff=cutoff)

                positions_motif[0] = positions_motif[0] + region_coordinates_batch["ix"].values[0]

                positions.append(positions_motif)
                indices.append(torch.ones_like(scores_motif, dtype=torch.int) * motif_ix)
                scores.append(scores_motif)
                strands.append(strands_motif)

        # positions = [batch_dim (region and position), site]
        positions = torch.concat(positions, -1)
        positions = positions[0] * region_size + positions[1]
        indices = torch.concat(indices, -1)
        scores = torch.concat(scores, -1)
        strands = torch.concat(strands, -1)

        # sort by position
        sorted_idx = torch.argsort(positions)
        positions = positions[sorted_idx]
        indices = indices[sorted_idx]
        scores = scores[sorted_idx]
        strands = strands[sorted_idx]

        # calculate indptr
        indptr = ind2ptr(positions, region_size * len(region_coordinates))

        # store
        self.positions = positions.cpu().numpy()
        self.indptr = indptr.cpu().numpy()
        self.indices = indices.cpu().numpy()
        self.scores = scores.cpu().numpy()
        self.strands = strands.cpu().numpy()
        self.motifs = motifs

        return self


def divide_regions_in_batches(region_coordinates, batch_size=10):
    region_coordinates["len"] = region_coordinates["end"] - region_coordinates["start"]

    assert len(region_coordinates["len"].unique()) == 1, "All regions should have the same size"

    region_coordinates["cumlen"] = (region_coordinates["end"] - region_coordinates["start"]).cumsum()

    region_coordinates["batch"] = (region_coordinates["cumlen"] // batch_size).astype(int)

    region_coordinates["batch"] = (region_coordinates["batch"] - region_coordinates["batch"].min()).astype(int)

    region_coordinates["ix"] = np.arange(region_coordinates.shape[0])

    return region_coordinates


def read_pwms(pwms_file):
    pwms = {}
    motif = None
    motif_id = None
    with open(pwms_file) as f:
        for line in f:
            if line.startswith(">"):
                if motif is not None:
                    pwms[motif_id] = motif
                motif_id = line[1:].strip("\n")
                motif = []
            else:
                motif.append([float(x) for x in line.split("\t")])

    pwms = {motif_id: torch.tensor(pwm) for motif_id, pwm in pwms.items()}

    return pwms


def scan(onehot, pwm, cutoff=0):
    n = onehot.shape[-2]
    k = pwm.shape[-2]

    # forward strand
    positive = torch.zeros(((*onehot.shape[:-2], n - k + 1)), device=onehot.device)
    for i in range(k):
        # to save memory we do the matrix multiplication once per motif position
        # this does not cause a significant slowdown
        x = torch.matmul(onehot, pwm[[i]].T)
        positive += x[..., i : n - k + i + 1, 0]
    del x

    found_positive = positive >= cutoff
    scores_positive = positive[found_positive]
    positions_positive = torch.stack(torch.where(found_positive)).to(torch.int)

    # reverse (complement) strand
    onehot_comp = onehot[..., [3, 2, 1, 0]]
    pwm_rev = pwm.flip(0)
    negative = torch.zeros(((*onehot.shape[:-2], n - k + 1)), device=onehot.device)
    for i in range(k):
        x = torch.matmul(onehot_comp, pwm_rev[[i]].T)
        negative += x[..., i : n - k + i + 1, 0]
    del x

    found_negative = negative >= cutoff
    scores_negative = negative[found_negative]
    positions_negative = torch.stack(torch.where(found_negative)).to(torch.int)
    positions_negative[-1, :] = positions_negative[-1, :] + k - 1

    # return stacked score across forward or reverse strands
    return (
        torch.cat([scores_positive, scores_negative]),
        torch.cat([positions_positive, positions_negative], -1).to(torch.int),
        torch.cat(
            [
                torch.ones_like(scores_positive, dtype=torch.int8),
                -torch.ones_like(scores_negative, dtype=torch.int8),
            ]
        ),
    )


def create_onehot(seq):
    """
    Sequence contains integers 0 (A), 1 (C), 2 (G), 3 (T), and 4 (N)
    """
    return torch.tensor(np.eye(5, dtype=np.float32)[seq][:, :-1])


def digitize_sequence(sequence):
    translate_table = {
        "A": 0,
        "C": 1,
        "G": 2,
        "T": 3,
        "N": 4,
        "a": 0,
        "c": 1,
        "g": 2,
        "t": 3,
    }  # alphabetic order
    return np.array([translate_table[x] for x in sequence], dtype=np.int8)



import chromatinhd.data.transcriptome
import dataclasses
import torch
import chromatinhd.sparse


@dataclasses.dataclass
class Result:
    value: torch.Tensor

    def to(self, device):
        self.value = self.value.to(device)
        return self


class Transcriptome:
    def __init__(
        self,
        transcriptome: chromatinhd.data.transcriptome.Transcriptome,
        layer: str = None,
    ):
        if layer is None:
            X = transcriptome.X
        else:
            X = transcriptome.layers[layer]
        if chromatinhd.sparse.is_sparse(X):
            self.X = X.dense()
        elif torch.is_tensor(X):
            self.X = X
        else:
            self.X = torch.from_numpy(X)

    def load(self, minibatch):
        X = self.X[minibatch.cells_oi, :][:, minibatch.genes_oi]
        return Result(value=X)



import pandas as pd
import chromatinhd as chd
import numpy as np
import torch


def test_digitize_sequence():
    assert (
        chd.data.motifscan.motifscan.digitize_sequence("ACGTN")
        == np.array([0, 1, 2, 3, 4])
    ).all()


def test_create_onehot():
    assert (
        chd.data.motifscan.motifscan.create_onehot(
            chd.data.motifscan.motifscan.digitize_sequence("ACGTN")
        )
        == torch.tensor(
            [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1], [0, 0, 0, 0]]
        )
    ).all()


def test_scan():
    pwm = torch.tensor(
        [
            [1.0, 0, 0, 0],
            [0.0, 1, 0, 0],
            [0.0, 0, 1, 0],
        ]
    )
    tests = [
        {
            "sequence": "ACGT",
            "pwm": pwm,
            "expected_scores": torch.tensor([3.0, 3.0]),
            "expected_positions": torch.tensor([[0, 0], [0, 3]]),
            "expected_strands": torch.tensor([1, -1]),
        },
        {
            "sequence": "AAAA",
            "pwm": pwm,
            "expected_scores": torch.tensor([]),
            "expected_positions": torch.tensor([]).reshape(2, 0),
            "expected_strands": torch.tensor([]),
        },
        {
            "sequence": "CGT",
            "pwm": pwm,
            "expected_scores": torch.tensor([3.0]),
            "expected_positions": torch.tensor([[0], [2]]),
            "expected_strands": torch.tensor([-1]),
        },
        {
            "sequence": "CGTCGT",
            "pwm": pwm,
            "expected_scores": torch.tensor([2.0, 3.0, 3.0]),
            "expected_positions": torch.tensor([[0, 0, 0], [2, 2, 5]]),
            "expected_strands": torch.tensor([1, -1, -1]),
        },
    ]

    for test in tests:
        onehot = chd.data.motifscan.motifscan.create_onehot(
            chd.data.motifscan.motifscan.digitize_sequence(test["sequence"])
        )[None, ...]
        pwm = test["pwm"]

        scores, positions, strands = chd.data.motifscan.motifscan.scan(
            onehot, pwm, cutoff=1.5
        )

        assert torch.equal(scores, test["expected_scores"]), (test["sequence"], scores)
        assert torch.equal(positions, test["expected_positions"].to(torch.int))
        assert torch.equal(strands, test["expected_strands"].to(torch.int8))


class TestMotifscan:
    def test_simple(self, tmp_path):
        # create sequence
        sequence_1 = "CGTT"
        sequence_2 = "ACGA"
        sequence_3 = "ACAA"

        fa_file = f"{tmp_path}/test.fa"
        with open(fa_file, "w", encoding="utf8") as f:
            f.write(">chr1\n")
            f.write(sequence_1 + "\n")
            f.write(">chr2\n")
            f.write(sequence_2 + "\n")
            f.write(">chr3\n")
            f.write(sequence_3 + "\n")

        region_size = len(sequence_1)
        # create regions
        region_coordinates = pd.DataFrame(
            {
                "chrom": ["chr1", "chr2", "chr3"],
                "start": [0, 0, 0],
                "end": [region_size, region_size, region_size],
                "tss": [1, 1, 1],
                "strand": [1, 1, -1],
            }
        )
        regions = chd.data.regions.Regions.create(
            path=f"{tmp_path}/regions",
            coordinates=region_coordinates,
        )

        motifscan = chd.data.motifscan.Motifscan(f"{tmp_path}/motifscan")

        # create motifs
        pwms = {
            "A": torch.tensor(
                [
                    [1.0, 0, 0, 0],
                    [1.0, 0, 0, 0],
                ]
            ),
            "B": torch.tensor(
                [
                    [1.0, 0, 0, 0],
                    [0.0, 1, 0, 0],
                    [0.0, 0, 1, 0],
                ]
            ),
        }
        motifs = pd.DataFrame(
            {
                "motif": ["A", "B"],
            }
        ).set_index("motif")
        cutoffs = [1.5, 1.5]
        motifs["cutoff"] = cutoffs

        motifscan = motifscan.from_pwms(
            pwms=pwms,
            regions=regions,
            fasta_file=fa_file,
            cutoffs=motifs["cutoff"],
            path=f"{tmp_path}/motifscan",
        )

        assert np.array_equal(
            motifscan.positions,
            np.array([2, 3, 4, 7, 8, 10], dtype=int),
        )
        assert np.array_equal(
            motifscan.scores,
            np.array([3.0, 2.0, 3.0, 2.0, 2.0, 2.0]),
        )
        assert np.array_equal(
            motifscan.strands,
            np.array([-1, -1, 1, -1, 1, 1], dtype=np.int8),
        )
        assert len(motifscan.indptr) == (region_size * len(region_coordinates) + 1)

        # get all sites
        motifscan = motifscan.from_pwms(
            pwms=pwms,
            regions=regions,
            fasta_file=fa_file,
            cutoffs=-99999,
            path=f"{tmp_path}/motifscan",
        )
        assert len(motifscan.indptr) == (region_size * len(region_coordinates) + 1)



from setuptools import setup, Extension
import numpy

setup_args = dict(
    ext_modules=[
        Extension(
            "chromatinhd.models.pred.loader.fragments_helpers",
            ["src/chromatinhd/models/diff/loader/fragments_helpers.c"],
            include_dirs=[numpy.get_include()],
            py_limited_api=True,
        ),
        Extension(
            "chromatinhd.models.pred.loader.fragments_helpers",
            ["src/chromatinhd/models/pred/loader/fragments_helpers.c"],
            include_dirs=[numpy.get_include()],
            py_limited_api=True,
        ),
    ]
)
setup(**setup_args)
